from enum import Enum, auto
from typing import Optional

from pydantic import BaseModel, validator


class TransformationType(Enum):
    GATE_SCHEDULE_FLAG_VALUE_MAPPING = auto()
    GENERATOR_PRODUCTION_SCHEDULE_FLAG_VALUE_MAPPING = auto()
    PLANT_PRODUCTION_SCHEDULE_FLAG_VALUE_MAPPING = auto()
    GATE_OPENING_METER_TO_PERCENT = auto()
    TO_BOOL = auto()
    ZERO_IF_NOT_ONE = auto()
    MULTIPLY = auto()
    MULTIPLY_FROM_OFFSET = auto()
    ADD = auto()
    ADD_FROM_OFFSET = auto()
    DYNAMIC_ADD_FROM_OFFSET = auto()
    RESERVOIR_LEVEL_TO_VOLUME = auto()
    STATIC = auto()
    DYNAMIC_STATIC = auto()
    ONE_IF_TWO = auto()
    ADD_WATER_IN_TRANSIT = auto()


class Transformation(BaseModel):  # type: ignore[no-redef]  # mypy says kwargs is redefined on this line ¯\_(ツ)_/¯
    transformation: TransformationType
    kwargs: Optional[dict] = None

    @validator("transformation", pre=True)
    def to_type(cls, value):
        return TransformationType[value.upper()] if isinstance(value, str) else value

    def to_dict(self) -> dict:
        return {"transformation": self.transformation.name, "kwargs": self.kwargs}



import re
from uuid import uuid4


def replace_nordic_letters(input_string: str) -> str:
    return input_string
    # return (
    #     input_string.replace("æ", "a")
    #     .replace("Æ", "A")
    #     .replace("ø", "o")
    #     .replace("Ø", "O")
    #     .replace("å", "a")
    #     .replace("Å", "A")
    # )


def special_case_handle_gate_number(name: str) -> None:
    """Must handle any gate numbers above 1 if found in other sources than YAML"""
    # TODO: extend to handle special case if needed
    if re.search(pattern=r"L[2-9]", string=name):
        print_warning(f"Potential gate {name} not in YAML!")


def print_warning(s: str) -> None:
    """Adds some nice colors to the printed text :)"""
    print(f"\033[91m[WARNING] {s}\033[0m")


def unique_str() -> str:
    return str(uuid4())



LOG_LEVEL = "DEBUG"


class StatusEventConfig:
    EVENT_EXTERNAL_ID_PREFIX = "POWEROPS_PROCESS"
    EVENT_TYPE_PREFIX = "POWEROPS_PROCESS"
    EVENT_SOURCE = "CogShop"


class RelationshipsConfig:
    LABEL_PREFIX = "relationship_to."
    INCREMENTAL_MAPPING_LABEL = f"{LABEL_PREFIX}incremental_mapping_sequence"
    OBJECTIVE_SEQUENCE_LABEL = f"{LABEL_PREFIX}objective_sequence"
    SHOP_RUN_LOG_FILE_LABEL = f"{LABEL_PREFIX}log_file"



from __future__ import annotations

from cognite.powerops.config import BootstrapConfig

# TODO we probably don't need this, since configuration will be entirely local


class ConfigurationClient:
    def retrieve(self, external_id: str) -> BootstrapConfig:
        ...

    def update(self, configuration: BootstrapConfig) -> BootstrapConfig:
        ...

    def delete(self, configuration: BootstrapConfig) -> None:
        ...

    def copy(self, configuration: BootstrapConfig, external_id: str) -> BootstrapConfig:
        """Create a copy of an existing configuration with a new external_id"""



class CogShopError(Exception):
    """Base exception for all CogShop related errors"""

    pass


class CogWriterError(CogShopError):
    pass


class CogReaderError(CogShopError):
    pass


class CogShopConfigError(CogShopError):
    pass



from __future__ import annotations

import tempfile
from functools import cached_property
from typing import List, Optional

import arrow
from cognite.client import CogniteClient

from cognite.powerops.preprocessor import knockoff_logging as logging
from cognite.powerops.preprocessor.data_classes import CogShopCase, CogShopConfig
from cognite.powerops.preprocessor.data_classes.time_series_mapping import TimeSeriesMapping
from cognite.powerops.preprocessor.exceptions import CogReaderError
from cognite.powerops.preprocessor.utils import (
    ShopMetadata,
    arrow_to_ms,
    download_file,
    log_and_reraise,
    retrieve_yaml_file,
)

from .get_fdm_data import Case, get_case

logger = logging.getLogger(__name__)


class CogReader:
    """Read data from CDF and preprocess for SHOP"""

    def __init__(
        self,
        client: CogniteClient,
        fdm_space_external_id: str,
        fdm_case_external_id: str,
        fdm_model_version: Optional[int] = None,
    ) -> None:
        self.client: CogniteClient = client

        self.fdm_space_external_id = fdm_space_external_id
        self.fdm_case_external_id = fdm_case_external_id
        self.fdm_model_version = fdm_model_version

        self._tmp_dir = tempfile.TemporaryDirectory()
        logger.info(f"{self.__class__.__name__} initialized.")

    @property
    def file_external_id_prefix(self) -> str:
        return f"SHOP_{self.cog_shop_config.watercourse}_"

    @property
    def fdm_case(self) -> Case:
        try:
            return self._fdm_case
        except AttributeError:
            raise RuntimeError("Expected to to run `retrieve_case()` method first.") from None

    def retrieve_case(self) -> None:
        self._fdm_case = get_case(
            self.client,
            self.fdm_space_external_id,
            self.fdm_case_external_id,
            model_version=self.fdm_model_version,
        )

    @cached_property
    def cog_shop_config(self) -> CogShopConfig:
        config = CogShopConfig(
            watercourse=self.fdm_case.scenario.model_template.watercourse,
            starttime=self.fdm_case.start_time,
            endtime=self.fdm_case.end_time,
        )
        logger.debug(f"{config} loaded from event metadata.")
        return config

    @cached_property
    def cog_shop_case(self) -> CogShopCase:
        """
        Download and parse model yaml file from cdf and merge with commands.
        Instantiate a CogShopCase from the resulting dict.
        """
        case_dict = retrieve_yaml_file(
            self.client,
            file_external_id=self.fdm_case.scenario.model_template.model.file_external_id,
        )
        case_dict["commands"] = self.fdm_case.scenario.commands.commands

        cog_shop_case = CogShopCase.from_dict(case_dict)
        logger.debug("SHOP case loaded.")
        return cog_shop_case

    @cached_property
    def base_mapping(self) -> List[TimeSeriesMapping]:
        mappings = self.fdm_case.scenario.model_template.base_mappings.items
        return [TimeSeriesMapping.from_mapping_model(m) for m in mappings]

    @cached_property
    def incremental_mapping(self) -> Optional[List[TimeSeriesMapping]]:
        return []
        # TODO uncomment, SHOP needs this:
        # relationships = self.client.relationships.list(
        #     source_external_ids=[self.shop_event.external_id],  # BROKEN
        #     data_set_ids=self.shop_event.data_set_id,           # BROKEN
        #     target_types=["sequence"],
        #     labels=LabelFilter(contains_all=[RelationshipsConfig.INCREMENTAL_MAPPING_LABEL]),
        # )
        # if not relationships:
        #     logger.info("No relationships to 'incremental mapping'.")
        #     return None

        # if len(relationships) > 1:
        #     raise CogReaderError(
        #         f"Expected only one incremental mapping. Found {len(relationships)} relationships:"
        #         f" {[rel.external_id for rel in relationships]}"
        #     )
        # sequence_external_id = relationships[0].target_external_id
        # # shop_type = "incremental_mapping"  # TODO: assert?
        # sequence_rows = retrieve_sequence_rows_as_dicts(
        #     client=self.client, external_id=sequence_external_id
        # )
        # return [TimeSeriesMapping.from_dict(row) for row in sequence_rows]

    def get_extra_files_metadata(self) -> List[dict[str, str]]:
        files = self.client.files.list(
            external_id_prefix=self.file_external_id_prefix,
            metadata=ShopMetadata(type="extra_data"),
        )
        if files:
            return [self.file_metadata_to_dict(fmd) for fmd in files]
        else:
            return []

    @staticmethod
    def file_metadata_to_dict(file_metadata):
        md = {"id": file_metadata.id}
        if (xid := file_metadata.external_id) is not None:
            md["external_id"] = xid
        if (n := file_metadata.name) is not None:
            md["name"] = n
        if (dsid := file_metadata.data_set_id) is not None:
            md["data_set_id"] = dsid
        return md

    @cached_property
    def extra_file_paths(self) -> Optional[List[str]]:
        logger.debug("Looking for 'extra' file.")
        files = self.client.files.list(
            external_id_prefix=self.file_external_id_prefix,
            metadata=ShopMetadata(type="extra_data"),
        )
        if not files:
            logger.debug(f"No 'extra' file found with and prefix '{self.file_external_id_prefix}'.")
            return None
        logger.debug(f"Found {len(files)} 'extra' files: {[file.external_id for file in files]}.")
        return [download_file(client=self.client, file=file, download_directory=self._tmp_dir.name) for file in files]

    def get_mapping_files_metadata(self):
        files = self.client.files.list(
            external_id_prefix=self.file_external_id_prefix,
            metadata=ShopMetadata(type="water_value_cut_file_reservoir_mapping"),
        )
        if files:
            return [self.file_metadata_to_dict(fmd) for fmd in files]
        else:
            return []

    @cached_property
    def mapping_file_paths(self) -> Optional[List[str]]:
        logger.debug("Looking for 'reservoir mapping' file.")
        files = self.client.files.list(
            external_id_prefix=self.file_external_id_prefix,
            metadata=ShopMetadata(type="water_value_cut_file_reservoir_mapping"),
        )
        if not files:
            logger.debug(f"No 'reservoir mapping' file found with and prefix '{self.file_external_id_prefix}'.")
            return None
        logger.debug(f"Found {len(files)} 'reservoir mapping' files: {[file.external_id for file in files]}.")
        return [download_file(client=self.client, file=file, download_directory=self._tmp_dir.name) for file in files]

    def get_cut_file_metadata(self):
        files = self.client.files.list(
            external_id_prefix=self.file_external_id_prefix,
            metadata=ShopMetadata(type="water_value_cut_file", watercourse=self.cog_shop_config.watercourse),
            limit=None,
        )
        if not files:
            logger.debug(f"No 'cut' file found with prefix '{self.file_external_id_prefix}'.")
            return None

        # Select file that is closest in time before starttime
        best_diff = float("inf")
        closest_file = None
        for this_file in files:
            try:
                # Assume datetime string after last "_"
                updated_at = this_file.metadata.get("update_datetime", this_file.external_id.split("_")[-1])
                updated_at_ms = arrow_to_ms(arrow.get(updated_at))  # parse ISO 8601 compliant string
                this_diff = self.cog_shop_config.starttime_ms - updated_at_ms

                if this_diff >= 0 and this_diff < best_diff:
                    logger.debug(f"Cutfile {this_file.external_id} is closer to starttime.")
                    best_diff = this_diff
                    closest_file = this_file

            except arrow.parser.ParserError as e:
                logger.warning(f"Failed to parse '{this_file.external_id}': {e}")
        if not closest_file:
            if len(files) > 0:
                logger.warning(
                    f"Could not find a cut file with a valid datetime - using {files[0].external_id}"
                    f" (which may be outdated)"
                )
                return self.file_metadata_to_dict(files[0])
            raise CogReaderError("Could not find any cut file!")

        return self.file_metadata_to_dict(closest_file)

    @cached_property
    def cut_file_path(self) -> Optional[str]:
        # TODO: logic for selecting cut file might need to be revised
        #   if getting additonal metadata about which date/time period the file is valid for.
        logger.debug("Looking for 'cut' file.")
        files = self.client.files.list(
            external_id_prefix=self.file_external_id_prefix,
            metadata=ShopMetadata(type="water_value_cut_file", watercourse=self.cog_shop_config.watercourse),
            limit=None,
        )
        if not files:
            logger.debug(f"No 'cut' file found with prefix '{self.file_external_id_prefix}'.")
            return None

        # Select file that is closest in time before starttime
        best_diff = float("inf")
        closest_file = None
        for this_file in files:
            try:
                # Assume datetime string after last "_"
                updated_at = this_file.metadata.get("update_datetime", this_file.external_id.split("_")[-1])
                updated_at_ms = arrow_to_ms(arrow.get(updated_at))  # parse ISO 8601 compliant string
                this_diff = self.cog_shop_config.starttime_ms - updated_at_ms

                if this_diff >= 0 and this_diff < best_diff:
                    logger.debug(f"Cutfile {this_file.external_id} is closer to starttime.")
                    best_diff = this_diff
                    closest_file = this_file

            except arrow.parser.ParserError as e:
                logger.warning(f"Failed to parse '{this_file.external_id}': {e}")
        if not closest_file:
            if len(files) > 0:
                logger.warning(
                    f"Could not find a cut file with a valid datetime - using {files[0].external_id}"
                    f" (which may be outdated)"
                )
                return download_file(client=self.client, file=files[0], download_directory=self._tmp_dir.name)
            raise CogReaderError("Could not find any cut file!")

        return download_file(client=self.client, file=closest_file, download_directory=self._tmp_dir.name)

    @cached_property
    def license_file_path(self) -> Optional[str]:
        """Read SHOP license file from CDF, store locally"""
        logger.debug("Looking for license file.")
        files = self.client.files.list(metadata=ShopMetadata(type="license"))
        if not files:
            logger.warning("Did not find any license file in CDF.")
            return None
        if len(files) > 1:
            logger.warning(f"Found more than one license file: {files}. Will use {files[0]}.")
        file_name = files[0].name
        if file_name != "SHOP_license.dat":
            raise CogReaderError(f"License file has to be named 'SHOP_license.dat'! Given file is named '{file_name}'.")
        download_file(client=self.client, file=files[0], download_directory=self._tmp_dir.name)
        return self._tmp_dir.name  # path to folder with license file instead of path to file

    @log_and_reraise(CogReaderError)
    def _replace_model_time_series(self) -> None:
        start = self.cog_shop_config.starttime_ms
        end = self.cog_shop_config.endtime_ms
        logger.debug("Applying 'base mapping'")
        self.cog_shop_case.replace_time_series(client=self.client, start=start, end=end, mappings=self.base_mapping)
        if self.incremental_mapping:
            logger.debug("Applying 'incremental mapping'")
            self.cog_shop_case.replace_time_series(
                client=self.client,
                start=start,
                end=end,
                mappings=self.incremental_mapping,
                dynamic_minute_offset=self.cog_shop_config.dynamic_minute_offset,
            )

    def run(self) -> CogReader:
        self.retrieve_case()

        # If specified in CogShopConfig (overrides previously set time specification)
        if self.cog_shop_config.timeresolution:
            self.cog_shop_case.set_time(
                starttime=self.cog_shop_config.starttime,
                endtime=self.cog_shop_config.endtime,
                timeresolution=self.cog_shop_config.timeresolution,
            )

        # Use defaults if time specification not already set
        if self.cog_shop_case.time is None:
            self.cog_shop_case.set_default_time(
                starttime=self.cog_shop_config.starttime, endtime=self.cog_shop_config.endtime
            )

        # TODO: consider if needed or commands should be required
        if self.cog_shop_case.commands is None:
            self.cog_shop_case._set_some_commands()

        self._replace_model_time_series()
        return self



from cognite.client.data_classes import Asset


def create_skeleton_asset_hierarchy(
    shop_service_url: str,
    organization_subdomain: str,
    tenant_id: str,
) -> list[Asset]:
    """Creates a skeleton Asset hierarchy for all Assets used by PowerOps"""
    root_external_id = "power_ops"
    configurations_external_id = "configurations"

    if shop_service_url == "https://shop-staging.az-inso-powerops.cognite.ai/submit-run":
        customer = "cognite"
    else:
        customer = organization_subdomain

    configurations_metadata = {
        "shop_service_url": shop_service_url,
        "organization_subdomain": organization_subdomain,
        "customer": customer,
        "tenant_id": tenant_id,
    }
    return [
        Asset(
            external_id=root_external_id,
            name="PowerOps",
        ),
        Asset(
            external_id="price_areas",
            name="Price areas",
            parent_external_id=root_external_id,
        ),
        Asset(
            external_id="watercourses",
            name="Watercourses",
            parent_external_id=root_external_id,
        ),
        Asset(
            external_id="plants",
            name="Plants",
            parent_external_id=root_external_id,
        ),
        Asset(
            external_id="reservoirs",
            name="Reservoirs",
            parent_external_id=root_external_id,
        ),
        Asset(
            external_id="generators",
            name="Generators",
            parent_external_id=root_external_id,
        ),
        Asset(
            external_id=configurations_external_id,
            name="Configurations",
            description="Configurations used for PowerOps",
            parent_external_id=root_external_id,
            metadata=configurations_metadata,
        ),
        Asset(
            external_id="benchmarking_configurations",
            name="Benchmarking configurations",
            parent_external_id=configurations_external_id,
            description="Configurations used in benchmarking processes",
        ),
        Asset(
            external_id="bid_process_configurations",
            name="Bid configurations",
            parent_external_id=configurations_external_id,
            description="Configurations used in bid matrix generation processes",
        ),
        Asset(
            external_id="rkom_bid_process_configurations",
            name="RKOM bid configurations",
            parent_external_id=configurations_external_id,
            description="Configurations used in RKOM bid generation processes",
        ),
        Asset(
            external_id="rkom_bid_combination_configurations",
            name="RKOM bid combination configurations",
            parent_external_id=configurations_external_id,
            description="Configurations for which bids should be combined into a total RKOM bid form",
        ),
        Asset(
            external_id="market_configurations",
            name="Market configurations",
            parent_external_id=configurations_external_id,
            description="Configurations used for different markets",
        ),
    ]



import os
import time
from datetime import datetime
from functools import wraps
from typing import Callable, Dict, Iterable, List, Optional, Type, Union

import arrow
import numpy as np
import pandas as pd
import yaml
from cognite.client import ClientConfig, CogniteClient
from cognite.client.credentials import OAuthClientCredentials, Token
from cognite.client.data_classes import Asset, Event, FileMetadata, LabelDefinition, Relationship, Sequence, TimeSeries
from cognite.client.exceptions import CogniteDuplicatedError, CogniteException

from cognite.powerops.preprocessor import knockoff_logging as logging

logger = logging.getLogger(__name__)


def ms_to_datetime(timestamp: int):
    """Milliseconds since Epoch to datetime."""
    return arrow.get(timestamp).datetime.replace(tzinfo=None)  # TODO: take tz into account


def arrow_to_ms(arrow: arrow.Arrow) -> int:
    """Arrow datetime to milliseconds since Epoch."""
    return int(arrow.float_timestamp * 1000)


def now() -> int:
    """Current time in milliseconds since Epoch"""
    return arrow_to_ms(arrow.now())


def datetime_from_str(dt: str) -> datetime:
    return arrow.get(dt).datetime.replace(tzinfo=None)


def shift_datetime_str(dt: str, days: int) -> str:
    return arrow.get(dt).shift(days=days).format("YYYY-MM-DD HH:mm:ss")


def get_cdf_client(
    project: str,
    tenant_id: str,
    client_id: str,
    client_secret: str,
    cluster: str,
) -> CogniteClient:
    """
    Return authenticated CogniteClient.
    """
    base_url = f"https://{cluster}.cognitedata.com"
    creds = OAuthClientCredentials(
        token_url=f"https://login.microsoftonline.com/{tenant_id}/oauth2/v2.0/token",
        client_id=client_id,
        client_secret=client_secret,
        scopes=[f"{base_url}/.default"],
    )

    return CogniteClient(
        config=ClientConfig(
            client_name="python-dd-local",
            base_url=base_url,
            project=project,
            credentials=creds,
        )
    )


def initialize_cognite_client() -> CogniteClient:
    logger.info("Setting up CogniteClient...")
    os.environ["COGNITE_DISABLE_PYPI_VERSION_CHECK"] = "true"

    token = os.getenv("TOKEN")
    project = os.getenv("PROJECT")
    base_url = os.getenv("BASE_URL") or "https://api.cognitedata.com"

    client = CogniteClient(
        config=ClientConfig(
            client_name="CogShop-local-debug",
            base_url=base_url,
            project=project,
            credentials=Token(token),
            timeout=90,
        )
    )

    logger.info(f"Client successfully set up towards '{project}'.")
    return client


def retrieve_yaml_file(client: CogniteClient, file_external_id: str) -> dict:
    """Return dict from YAML file in CDF"""
    f = client.files.download_bytes(external_id=file_external_id)
    return yaml.safe_load(f)


def save_dict_as_yaml(file_path: str, d: dict) -> None:
    with open(file_path, "w") as file:
        yaml.dump(d, file, allow_unicode=True)


def dict_to_yaml_string(d: dict) -> str:
    return yaml.dump(d, allow_unicode=True)


def rename_dict_keys(d: Dict, key_mapping: Dict) -> None:
    for old_key, new_key in key_mapping.items():
        d[new_key] = d.pop(old_key)


def log_and_reraise(exception_to_raise: Type[Exception]):
    """Wrapper that logs any exceptions and reraising the specified exception to raise"""

    def _log_and_raise(func: Callable):
        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as e:
                msg = f"{func.__name__} failed!"
                logger.error(msg)
                raise exception_to_raise(msg) from e

        return wrapper

    return _log_and_raise


def _overlapping_keys(dict_: dict, other: dict) -> List[str]:
    return list(set(dict_.keys()).intersection(other.keys()))


def merge_dicts(*args: dict) -> dict:
    merged: dict = {}
    for dict_ in args:
        assert isinstance(dict_, dict), f"Expected dict, got {type(dict_)}!"
        overlap = _overlapping_keys(dict_, merged)
        if overlap:
            raise Exception(f"Key collision on '{overlap}' when merging dictionaries!")
        merged = {**merged, **dict_}
    return merged


def simple_relationship(
    source: Union[Asset, TimeSeries, FileMetadata, Sequence, Event],
    target: Union[Asset, TimeSeries, FileMetadata, Sequence, Event],
    label_external_id: str,
    data_set_id: int,
) -> Relationship:
    """Simplifies Cognite Python SDK creation of Relationships."""
    external_id = f"{source.external_id}.{target.external_id}"
    source_type = source.__class__.__name__ if not isinstance(source, FileMetadata) else "file"
    target_type = target.__class__.__name__ if not isinstance(target, FileMetadata) else "file"

    return Relationship(
        external_id=external_id,
        source_type=source_type,
        target_type=target_type,
        source_external_id=source.external_id,
        target_external_id=target.external_id,
        data_set_id=data_set_id,
        labels=[label_external_id],
    )


# TODO: should Label creation be handled here?
def create_relationship(
    client: CogniteClient, source, target, label_external_id: str, data_set_id: int
) -> Relationship:
    try:
        relationship = simple_relationship(source, target, label_external_id, data_set_id)
        client.relationships.create(relationship)
    except CogniteException:
        logger.warning("Failed to create Relationship! Will try to create LabelDefinition first and try again.")
        create_label_definition(
            client=client,
            label_external_id=label_external_id,
            data_set_id=relationship.data_set_id,
        )
        time.sleep(1)
        client.relationships.create(relationship=relationship)
    logger.debug(
        f"Created relationship between {source.__class__.__name__}({source.external_id}) and"
        f" {target.__class__.__name__}({target.external_id})"
    )
    return relationship


def create_label_definition(client: CogniteClient, label_external_id: str, data_set_id: int) -> None:
    try:
        label_definition = LabelDefinition(
            external_id=label_external_id, name=label_external_id, data_set_id=data_set_id
        )
        client.labels.create(label=label_definition)
        logger.debug(f"Created LabelDefinition({label_definition.external_id})")
    except CogniteDuplicatedError:
        logger.debug(f"LabelDefinition({label_definition.external_id}) already exists.")


def retrieve_sequence_rows_as_dicts(client: CogniteClient, external_id: str) -> List[dict]:
    df = client.sequences.data.retrieve_dataframe(external_id=external_id, start=0, end=None)
    df = df.where(pd.notnull(df), None)  # Replace NaNs with None
    return df.to_dict(orient="records")


# TODO: validation / enum
class ShopMetadata(dict):
    def __init__(
        self,
        type: Optional[str] = None,
        watercourse: Optional[str] = None,
        attribute: Optional[str] = None,
        object_type: Optional[str] = None,
        object_name: Optional[str] = None,
        penalty_breakdown: Optional[str] = None,
    ):
        super().__init__()
        # Iterate over all local variables with values
        vars = {attr: value for attr, value in locals().items() if not attr.startswith(("__"))}
        for attr, value in vars.items():
            if attr != "self" and value is not None:
                self[f"shop:{attr}"] = value


def download_file(client: CogniteClient, file: FileMetadata, download_directory: str) -> str:
    """Downloads file to the specified directory and returns the path of the downloaded file"""
    logger.debug(f"Downloading '{file.external_id}'.")
    client.files.download(directory=download_directory, external_id=file.external_id)
    file_path = f"{download_directory}/{file.name}"
    logger.debug(f"File saved to '{file_path}'")
    return file_path


# TODO: consider exception
def log_missing(expected: Iterable, actual: Iterable) -> None:
    if missing := set(expected).difference(actual):
        logger.warning(f"Missing: {missing}")


def remove_duplicates(lst: list) -> list:
    return list(set(lst))


# ! TODO: Confirm that None: True and "": True is OK
def retrieve_is_step(client: CogniteClient, mappings) -> Dict[Optional[str], bool]:
    """Create mapping between time series external_id and wether that time series is a step time series or not."""
    # Remove any duplicates and ignore None
    external_ids = list(
        {mapping.time_series_external_id for mapping in mappings if mapping.time_series_external_id is not None}
    )
    time_series = client.time_series.retrieve_multiple(external_ids=external_ids, ignore_unknown_ids=True)
    # NOTE: If no time series is specified (static mapping) it is assumed that the is_step=True
    return {None: True, "": True, **{ts.external_id: ts.is_step for ts in time_series}}


# TODO: refactor
# TODO: naming: time_series vs. datapoints
def retrieve_latest(client: CogniteClient, external_ids: List[Optional[str]], before: int) -> Dict[str, pd.Series]:
    if not external_ids:
        return {}
    external_ids = remove_duplicates(external_ids)
    logger.debug(f"Retrieving {external_ids} before '{ms_to_datetime(before)}'")
    time_series = client.datapoints.retrieve_latest(external_id=external_ids, before=before, ignore_unknown_ids=True)

    # For (Cog)Datapoints in (Cog)DatapointsList
    for datapoints in time_series:
        if len(datapoints) > 0:  # TODO: what to do about ts with no datapoints?
            datapoints.timestamp[0] = before  # align timestamps

    res = {
        datapoints.external_id: datapoints.to_pandas().iloc[:, 0]  # iloc to convert DataFrame to Series
        for datapoints in time_series
        if len(datapoints) > 0
    }
    log_missing(external_ids, res)
    return res


def _retrieve_range(client: CogniteClient, external_ids: List[str], start: int, end: int) -> pd.DataFrame:
    # TODO: Upgrade cognite-sdk to v5 (or later), and see how much of the code we can replace with direct SDK calls
    # - client.time_series.data.retrieve_dataframe(…, uniform_index=True) should give us almost what we want,
    # but maybe we need to be careful with cases where there is more than 1 hour between values
    # (I do not remember if this is an issue only for some aggregates like average, or for all).
    # And maybe we need to parametrise the "minimum resolution"
    # (seems to assume 1 hour, but we should support sub-hourly resolutuon)

    # Retrieve raw datapoints
    external_ids = remove_duplicates(external_ids)
    logger.debug(f"Retrieving {external_ids} between '{ms_to_datetime(start)}' and '{ms_to_datetime(end)}'")
    df_range = client.datapoints.retrieve(
        external_id=external_ids, start=start, end=end, ignore_unknown_ids=True
    ).to_pandas()

    # Retrieve latest datapoints before start
    df_latest = client.datapoints.retrieve_latest(
        external_id=external_ids, before=start, ignore_unknown_ids=True
    ).to_pandas()

    # Make sure we have a start timestamp in range
    if df_range.empty:
        df_range = pd.DataFrame(
            columns=df_latest.columns,
            index=pd.DatetimeIndex(data=np.array([start], dtype="datetime64[ms]")),
        )

    # Add the latest datapoints to the DataFrame
    df_raw = df_range.combine_first(df_latest)

    # Must retrieve time series metadata to correctly resample and aggregate datapoints
    time_series = client.time_series.retrieve_multiple(external_ids=external_ids, ignore_unknown_ids=True)
    step_columns = [ts.external_id for ts in time_series if ts.is_step]
    linear_columns = [ts.external_id for ts in time_series if not ts.is_step]
    logger.debug(f"time_series.is_step: True [{len(step_columns)}] False [{len(linear_columns)}]")

    # Step interpolation of time series with .is_step=False
    # NOTE: do not need to upsample when forward filling
    # TODO: note 2x ffill()
    df_step = df_raw[step_columns].ffill().resample("1h").ffill()

    # Linear interpolation of time series with .is_step=False
    # TODO: must upsample before downsampling?
    # TODO: confirm operations
    df_linear = df_raw[linear_columns].resample("1min").interpolate().resample("1h").interpolate()

    # Merge the step interpolated and linearly interpolated DataFrames
    df_combined = df_step.combine_first(df_linear)

    # Only return datapoints within the range
    df_filtered = df_combined[ms_to_datetime(start) : ms_to_datetime(end + 1)]

    log_missing(expected=external_ids, actual=df_filtered.columns)
    return df_filtered


def retrieve_range(client: CogniteClient, external_ids: List[str], start: int, end: int) -> Dict[str, pd.Series]:
    df = _retrieve_range(client=client, external_ids=external_ids, start=start, end=end)
    return {col: df[col].dropna() for col in df.columns}


def retrieve_time_series_datapoints(
    client: CogniteClient, mappings, start, end  # : List[TimeSeriesMapping]
) -> Dict[str, pd.Series]:
    time_series_start = retrieve_latest(
        client=client,
        external_ids=[mapping.time_series_external_id for mapping in mappings if mapping.retrieve == "START"],
        before=start,
    )
    time_series_end = retrieve_latest(
        client=client,
        external_ids=[mapping.time_series_external_id for mapping in mappings if mapping.retrieve == "END"],
        before=end,
    )
    time_series_range = retrieve_range(
        client=client,
        external_ids=[mapping.time_series_external_id for mapping in mappings if mapping.retrieve == "RANGE"],
        start=start,
        end=end,
    )
    _time_series_none = [mapping.shop_model_path for mapping in mappings if not mapping.retrieve]
    logger.debug(f"Not retrieving datapoints for {_time_series_none}")

    return merge_dicts(time_series_start, time_series_end, time_series_range)



from cognite.powerops.client import PowerOpsClient
from cognite.powerops.client.data_classes.case import Case
from cognite.powerops.version import __version__

__all__ = ["__version__", "Case", "PowerOpsClient"]



from functools import cached_property
from typing import Optional

from cognite.client import ClientConfig, CogniteClient
from cognite.dm_clients.config import settings

from cognite.powerops.client.api.asset_apis import (
    BenchmarkingConfigurationsAPI,
    BidConfigurationsAPI,
    GeneratorsAPI,
    MarketConfigurationsAPI,
    PlantAPI,
    PriceAreasAPI,
    ReservoirsAPI,
    RKOMBidCombinationConfiguration,
    RKOMBidConfigurationsAPI,
    WatercourseAPI,
)
from cognite.powerops.client.api.dm_apis import CaseAPI, CommandsAPI, MappingAPI, ScenarioAPI, TransformationAPI
from cognite.powerops.client.api.shop_api import ShopAPI
from cognite.powerops.client.config_client import ConfigurationClient
from cognite.powerops.client.dm.client import get_power_ops_dm_client
from cognite.powerops.utils.cdf_utils import retrieve_dataset


class ConfigurationsClient:
    def __init__(self, read_dataset: str, write_dataset: str, client: CogniteClient):
        self.bids = BidConfigurationsAPI(client, read_dataset, write_dataset)
        self.rkom_bids = RKOMBidConfigurationsAPI(client, read_dataset, write_dataset)
        self.bechmarkings = BenchmarkingConfigurationsAPI(client, read_dataset, write_dataset)
        self.markets = MarketConfigurationsAPI(client, read_dataset, write_dataset)
        self.rkom_bid_combinations = RKOMBidCombinationConfiguration(client, read_dataset, write_dataset)


class PowerOpsClient:
    def __init__(
        self,
        read_dataset: Optional[str] = None,
        write_dataset: Optional[str] = None,
        cogshop_version: Optional[str] = None,
        config: Optional[ClientConfig] = None,
        space_id: Optional[str] = None,
        data_model: Optional[str] = None,
        schema_version: Optional[int] = None,
    ):
        self._read_dataset = read_dataset or settings.powerops.read_dataset
        self._write_dataset = write_dataset or settings.powerops.write_dataset
        self._cogshop_version = cogshop_version or settings.powerops.get("cogshop_version", "")

        self.dm = get_power_ops_dm_client(
            config=config, space_id=space_id, data_model=data_model, schema_version=schema_version
        )
        self.cdf = self.dm._client

        self.configurations = ConfigurationClient()

        self.shop = ShopAPI(po_client=self)

        self.configurations = ConfigurationsClient(read_dataset, write_dataset, self.cdf)
        self.generators = GeneratorsAPI(self.cdf, read_dataset, write_dataset)
        self.plants = PlantAPI(self.cdf, read_dataset, write_dataset)
        self.price_areas = PriceAreasAPI(self.cdf, read_dataset, write_dataset)
        self.reservoirs = ReservoirsAPI(self.cdf, read_dataset, write_dataset)
        self.watercourses = WatercourseAPI(self.cdf, read_dataset, write_dataset)

        self.cases = CaseAPI(self.dm)
        self.commands = CommandsAPI(self.dm)
        self.scenarios = ScenarioAPI(self.dm)
        self.mappings = MappingAPI(self.dm)
        self.transformations = TransformationAPI(self.dm)

    @cached_property
    def read_dataset_id(self) -> int:
        return retrieve_dataset(self.cdf, external_id=self._read_dataset).id

    @cached_property
    def write_dataset_id(self) -> int:
        return retrieve_dataset(self.cdf, external_id=self._write_dataset).id

    @property
    def cogshop_version(self) -> str:
        return self._cogshop_version



import os
from typing import Optional
from uuid import uuid4

import yaml
from cognite.client import CogniteClient

from cognite.powerops.preprocessor import knockoff_logging as logging
from cognite.powerops.preprocessor.cogreader import CogReader
from cognite.powerops.preprocessor.utils import initialize_cognite_client

logger = logging.getLogger(__name__)


def _main(
    client: CogniteClient,
    fdm_space_external_id: str,
    fdm_case_external_id: str,
    output_data_set_id: int,
    fdm_model_version: Optional[int] = None,
) -> dict:
    cog_reader = CogReader(
        client=client,
        fdm_space_external_id=fdm_space_external_id,
        fdm_case_external_id=fdm_case_external_id,
        fdm_model_version=fdm_model_version,
    ).run()

    cog_shop_case = cog_reader.cog_shop_case.to_dict()
    cog_shop_case_yaml = yaml.dump(cog_shop_case, allow_unicode=True, encoding="utf-8")

    extra_files_md = cog_reader.get_extra_files_metadata()
    mapping_files_md = cog_reader.get_mapping_files_metadata()
    cut_file_md = cog_reader.get_cut_file_metadata()

    xid = f"cog_shop_preprocessor/{fdm_space_external_id}/{fdm_case_external_id}/{str(uuid4())}"

    cog_shop_case_file_md = client.files.upload_bytes(
        content=cog_shop_case_yaml,
        name=f"{fdm_space_external_id}-{fdm_case_external_id}-case.yaml",
        mime_type="application/yaml",
        external_id=xid,
        metadata={
            "fdm:space_external_id": fdm_space_external_id,
            "fdm:case_external_id": fdm_case_external_id,
            "shop:type": "cog_shop_case",
        },
        data_set_id=output_data_set_id,
    )

    return {
        "cog_shop_case_file": cog_reader.file_metadata_to_dict(cog_shop_case_file_md),
        "cut_file": cut_file_md,
        "mapping_files": mapping_files_md,
        "extra_files": extra_files_md,
    }


def main(
    output_data_set_id: int,
    client: CogniteClient = None,
    fdm_space_external_id: Optional[str] = None,
    fdm_case_external_id: Optional[str] = None,
    fdm_model_version: Optional[int] = None,
) -> dict:
    client = client if client is not None else initialize_cognite_client()

    fdm_space_external_id = (
        fdm_space_external_id if fdm_space_external_id is not None else os.getenv("FDM_SPACE_EXTERNAL_ID")
    )
    fdm_case_external_id = (
        fdm_case_external_id if fdm_case_external_id is not None else os.getenv("FDM_CASE_EXTERNAL_ID")
    )

    result = _main(
        client=client,
        fdm_space_external_id=fdm_space_external_id,  # type: ignore[arg-type]
        fdm_case_external_id=fdm_case_external_id,  # type: ignore[arg-type]
        output_data_set_id=output_data_set_id,
        fdm_model_version=fdm_model_version,
    )

    logger.info(f"main() returned: {result!r}")

    return result



from datetime import datetime
from pathlib import Path
from typing import Optional

import yaml
from cognite.client import CogniteClient

from cognite.powerops.data_classes.shop_file_config import ShopFileConfig
from cognite.powerops.utils.serializer import load_yaml


def upload_shop_config_file(
    client: CogniteClient,
    config: ShopFileConfig,
    data_set_id: int,
    overwrite: bool = True,
) -> None:
    if config.path is None:
        raise ValueError("The Path must be set to upload file")
    file_content = Path(config.path).read_bytes()
    if config.md5_hash is None:
        config.set_md5_hash(file_content)
    try:
        file = client.files.upload_bytes(
            content=file_content,
            external_id=config.external_id,
            name=config.external_id,
            metadata=config.metadata,
            source="PowerOps bootstrap",
            mime_type="text/plain",
            data_set_id=data_set_id,
            overwrite=overwrite,
        )
        print(f"Uploaded file with externalId {file.external_id}")
    except Exception as e:
        print(e)


def shop_attribute_value_is_time_series(shop_attribute_value) -> bool:
    return isinstance(shop_attribute_value, dict) and isinstance(list(shop_attribute_value)[0], datetime)


# ! Assumes yaml with "model", "time" and "connections"!!
# --> TODO: "time" is just dropped?
# TODO: extract loading of yaml
def get_model_without_timeseries(
    yaml_raw_path: str,
    non_time_series_attributes_to_remove: Optional[list[str]] = None,
    encoding=None,
) -> dict:
    if non_time_series_attributes_to_remove is None:
        non_time_series_attributes_to_remove = []

    model_incl_time_and_connections = load_yaml(Path(yaml_raw_path), encoding=encoding, clean_data=True)

    model = model_incl_time_and_connections["model"]

    for object_type, objects in model.items():
        for object_attributes in objects.values():
            for attribute_name, attribute_data in list(
                object_attributes.items()
            ):  # Needs list() since we pop while iterating over the dict
                if f"{object_type}.{attribute_name}" in non_time_series_attributes_to_remove:
                    object_attributes.pop(attribute_name)

                elif shop_attribute_value_is_time_series(attribute_data):
                    object_attributes.pop(attribute_name)

    model_dict = {"model": model}

    # TODO: remove this when standardizing input SHOP file strucutre (model vs model + connections)
    if "connections" in model_incl_time_and_connections:
        model_dict["connections"] = model_incl_time_and_connections["connections"]

    return model_dict


def process_yaml_file(yaml_raw_path: str, yaml_processed_path: str) -> None:
    """Process raw YAML file and store as new file"""
    # Remove timeseries
    model_without_timeseries = get_model_without_timeseries(
        yaml_raw_path=yaml_raw_path,
        non_time_series_attributes_to_remove=["reservoir.start_vol", "generator.initial_state"],
        encoding="utf-8",
    )

    # Save cleaned model
    with open(yaml_processed_path, "w", encoding="utf-8") as stream:
        yaml.safe_dump(data=model_without_timeseries, stream=stream, allow_unicode=True, sort_keys=False)



from cognite.powerops.client.data_classes.case import Case
from cognite.powerops.client.data_classes.shop_result_files import ShopLogFile, ShopResultFile, ShopYamlFile
from cognite.powerops.client.data_classes.shop_results import ObjectiveFunction, ShopRunResult
from cognite.powerops.client.data_classes.shop_run import ShopRun
from cognite.powerops.client.data_classes.shop_run_event import ShopRunEvent

__all__ = [
    "Case",
    "ObjectiveFunction",
    "ShopRunResult",
    "ShopLogFile",
    "ShopResultFile",
    "ShopRun",
    "ShopRunEvent",
    "ShopYamlFile",
]



from __future__ import annotations

from typing import Generic, TypeVar

from cognite.dm_clients.domain_modeling import DomainModel, DomainModelAPI

from cognite.powerops.client.dm.client import PowerOpsDmClient
from cognite.powerops.client.dm.schema import Case, CommandsConfig, Mapping, Scenario, Transformation

DomainModelT = TypeVar("DomainModelT", bound=DomainModel)


class DMAPIBase(Generic[DomainModelT]):
    model_class: DomainModelT
    dm_attr: str

    def __init__(self, dm_client: PowerOpsDmClient):
        self.dm_client = dm_client

    @property
    def model_api(self) -> DomainModelAPI[DomainModelT]:
        return getattr(self.dm_client, self.dm_attr)

    # CRUD

    def retrieve(self, external_id: str) -> DomainModelT:
        items = self.model_api.retrieve([external_id])
        if len(items) < 1:
            raise ValueError(f"{self.model_class.__name__} with externalId {external_id}" f" not found.")
        if len(items) > 1:
            # this should never happen
            raise ValueError(f"Multiple {self.model_class.__name__} items with externalId" f" {external_id} found.")
        return items[0]

    def update(self, transformation: DomainModelT) -> DomainModelT:
        self.model_api.apply([transformation])
        return transformation

    def delete(self, transformation: DomainModelT) -> None:
        self.model_api.delete([transformation])


class CaseAPI(DMAPIBase[Case]):
    model_class = Case
    dm_attr = "case"


class CommandsAPI(DMAPIBase[CommandsConfig]):
    model_class = CommandsConfig
    dm_attr = "commands"


class MappingAPI(DMAPIBase[Mapping]):
    model_class = Mapping
    dm_attr = "mapping"


class ScenarioAPI(DMAPIBase[Scenario]):
    model_class = Scenario
    dm_attr = "scenario"


class TransformationAPI(DMAPIBase[Transformation]):
    model_class = Transformation
    dm_attr = "transformation"



import json
from enum import Enum, auto
from typing import Dict, List, Optional, Tuple, Union

from pydantic import BaseModel, validator

from cognite.powerops.logger import LoggingLevelT


class CommonConstants(BaseModel):
    data_set_external_id: str
    overwrite_data: bool
    organization_subdomain: str
    tenant_id: str
    shop_version: str
    skip_dm: bool = False
    debug_level: LoggingLevelT = "INFO"


class RelativeTime(BaseModel):
    relative_time_string: Optional[str]
    operations: Optional[List[Tuple[str, Union[str, Dict[str, int]]]]]

    @validator("operations", pre=True, always=True)
    def to_old_format(cls, value):
        if not isinstance(value, list):
            return value

        old_formats = []
        for v in value:
            if isinstance(v, dict):
                operation, argument = next(iter(v.items()))
                old_formats.append((operation, argument))
            else:
                # Already old format
                old_formats.append(v)
        return old_formats

    @validator("operations", pre=True, always=True)
    @classmethod
    def _parse_relative_time_string(cls, operations, values):
        # NOTE: tuples will be parsed as lists when dumping to string
        if operations:
            return operations
        elif values["relative_time_string"] == "tomorrow":
            return [("shift", {"days": 1}), ("floor", "day")]
        elif values["relative_time_string"] == "end_of_next_week":
            return [("floor", "week"), ("shift", {"weeks": 2})]
        elif values["relative_time_string"] == "monday":
            # Monday next week (given that we are before friday 12:00)
            return [("shift", {"hours": 12}), ("shift", {"weekday": 4}), ("floor", "day"), ("shift", {"weekday": 0})]
        elif values["relative_time_string"] == "saturday":
            # This Saturday (given that we are before thursday 12:00)
            return [("shift", {"hours": 12}), ("shift", {"weekday": 3}), ("floor", "day"), ("shift", {"weekday": 5})]
        else:
            raise ValueError(f"{values['relative_time_string']} not a valid value for relative_time_string")

    def __str__(self) -> str:
        return json.dumps(self.operations)


class RetrievalType(Enum):
    RANGE = auto()
    START = auto()
    END = auto()


class AggregationMethod(Enum):
    sum = auto()
    mean = auto()
    std = auto()
    sem = auto()
    max = auto()
    min = auto()
    median = auto()
    first = auto()
    last = auto()



from __future__ import annotations

import logging
import tempfile
from pathlib import Path
from typing import TYPE_CHECKING, Any, Generic, NewType, Optional, Sequence, TextIO, Tuple, TypeVar, Union, cast

import yaml
from cognite.client.data_classes import FileMetadata

from cognite.powerops.client.data_classes import ShopLogFile, ShopResultFile, ShopYamlFile
from cognite.powerops.utils.cdf_utils import retrieve_relationships_from_source_ext_id

if TYPE_CHECKING:
    from cognite.powerops import PowerOpsClient

logger = logging.getLogger(__name__)


ShopResultFileT = TypeVar("ShopResultFileT", bound=ShopResultFile)
EncodingT = NewType("EncodingT", str)


class ShopFilesAPI:
    def __init__(self, po_client: PowerOpsClient) -> None:
        self._po_client = po_client
        self.log_files = ShopLogFilesAPI(po_client)
        self.yaml_files = ShopYamlFilesAPI(po_client)

    def download_to_disk(self, shop_file_id: str, dir_path: Path) -> None:
        """Download a file from CDF to local filesystem."""
        self._po_client.cdf.files.download_to_path(path=dir_path / shop_file_id, external_id=shop_file_id)

    def retrieve_related_files_metadata(
        self, source_external_id: str, label_ext_id: Optional[Union[str, Sequence[str]]] = None
    ) -> Sequence[FileMetadata]:
        """
        Retrieve metadata of files that have a relationship to specified source (externalId).
        Optionally restrict the results by relationship labels.
        """
        relationships = retrieve_relationships_from_source_ext_id(
            self._po_client.cdf,
            source_ext_id=source_external_id,
            label_ext_id=label_ext_id,
            target_types=["file"],
        )
        if not relationships:
            return []
        return self._po_client.cdf.files.retrieve_multiple(
            external_ids=[rel.target_external_id for rel in relationships],
            ignore_unknown_ids=True,
        )


class ShopResultFilesAPI(Generic[ShopResultFileT]):
    def __init__(self, po_client: PowerOpsClient) -> None:
        self._po_client = po_client

    def retrieve(self, file_metadata: FileMetadata) -> ShopResultFileT:
        raise NotImplementedError()

    def _parse_file(self, file: TextIO) -> Any:
        raise NotImplementedError()

    def _download(self, file_metadata: FileMetadata) -> Tuple[Any, EncodingT]:
        """Download the file, parse it and return the content and encoding used to open the file."""
        encoding: EncodingT = file_metadata.metadata.get("encoding", "utf-8")
        with tempfile.TemporaryDirectory() as tmp_dir:
            self._po_client.shop.files.download_to_disk(file_metadata.external_id, Path(tmp_dir))
            tmp_path = Path(tmp_dir) / file_metadata.external_id
            content_and_encoding: Optional[Tuple[Any, EncodingT]]
            try:
                with open(tmp_path, "r", encoding=encoding) as downloaded_file:
                    content_and_encoding = self._parse_file(downloaded_file), encoding
            except UnicodeDecodeError:
                content_and_encoding = self._download_w_wrong_encoding(encoding, tmp_path)
                if content_and_encoding is None:
                    raise
            return content_and_encoding

    def _download_w_wrong_encoding(self, encoding: EncodingT, tmp_path: Path) -> Optional[Tuple[Any, EncodingT]]:
        """
        Some files are uploaded to CDF with latin-1 encoding, most with utf-8.
        When utf-8 fails, we try latin-1.
        """
        if encoding != "latin1":
            encoding = cast(EncodingT, "latin-1")
            with open(tmp_path, "r", encoding=encoding) as downloaded_file:
                try:
                    return self._parse_file(downloaded_file), encoding
                except (ValueError, TypeError):
                    pass
        return None


class ShopLogFilesAPI(ShopResultFilesAPI[ShopLogFile]):
    def retrieve(self, file_metadata: FileMetadata) -> ShopLogFile:
        content, encoding = self._download(file_metadata)
        return ShopLogFile(content, file_metadata, encoding)

    def _parse_file(self, file: TextIO) -> str:
        return file.read()


class ShopYamlFilesAPI(ShopResultFilesAPI[ShopYamlFile]):
    def retrieve(self, file_metadata: FileMetadata) -> ShopYamlFile:
        content, encoding = self._download(file_metadata)
        return ShopYamlFile(content, file_metadata, encoding)

    def _parse_file(self, file: TextIO) -> dict:
        return yaml.safe_load(file)



import json

import pandas as pd
from dotenv import dotenv_values
from yaml import safe_load

from cognite.powerops.preprocessor.utils import get_cdf_client


def make_base_mapping(row) -> dict:
    json_str = "".join(
        (
            row["transformations"],
            row["transformations1"],
            row["transformations2"],
            row["transformations3"],
        )
    )
    transformations = json.loads(json_str)

    return {
        "path": row["shop_model_path"],
        "timeseries_external_id": row["time_series_external_id"],
        "retrieve": row["retrieve"],
        "aggregation": row["aggregation"],
        "transformations": {
            "items": [
                {
                    "method": t["transformation"],
                    "arguments": json.dumps(t["kwargs"]) if t["kwargs"] is not None else None,
                }
                for t in transformations
            ]
        },
    }


env = dotenv_values(".env.heco-dev")

client = get_cdf_client(
    project=env["cdf_project"],
    tenant_id=env["cdf_tenant_id"],
    client_id=env["cdf_client_id"],
    client_secret=env["cdf_client_secret"],
    cluster=env["cdf_cluster"],
)


sequence_xid = "SHOP_Otta_base_mapping"
base_mapping = client.sequences.data.retrieve_dataframe(external_id=sequence_xid, limit=10000, start=0, end=-1)
base_mapping.where(pd.notnull(base_mapping), None, inplace=True)

mappings = []
for i, row in base_mapping.iterrows():
    mappings.append(make_base_mapping(row))


commands_file = client.files.download_bytes(external_id="SHOP_Glomma_commands")


# POWEROPS_SHOP_RUN_1b424f47-d893-425c-95e2-ed5b44c047c6

case = {
    "start_time": "2023-03-28 22:00:00",
    "end_time": "2023-04-09 22:00:00",
    "scenario": {
        "name": "Otta-Base",
        "commands": {"commands": safe_load(commands_file)["commands"]},
        "model_template": {
            "watercourse": "Otta",
            "model": {"file_external_id": "SHOP_Otta_model"},
            "base_mappings": {"items": mappings},
        },
    },
}



from __future__ import annotations

import os
from pathlib import Path
from typing import List, Literal, Optional

from cognite.powerops.config import (
    BenchmarkingConfig,
    BidMatrixGeneratorConfig,
    BidProcessConfig,
    BootstrapConfig,
    PriceScenario,
    RKOMBidCombinationConfig,
    RKOMBidProcessConfig,
    RkomMarketConfig,
    WatercourseConfig,
)
from cognite.powerops.data_classes.cdf_resource_collection import BootstrapResourceCollection
from cognite.powerops.data_classes.shop_file_config import ShopFileConfig, ShopFileConfigs
from cognite.powerops.data_classes.shop_output_definition import ShopOutputConfig
from cognite.powerops.data_classes.time_series_mapping import TimeSeriesMapping, write_mapping_to_sequence
from cognite.powerops.utils.cdf_auth import get_client
from cognite.powerops.utils.files import process_yaml_file
from cognite.powerops.utils.labels import create_labels
from cognite.powerops.utils.powerops_asset_hierarchy import create_skeleton_asset_hierarchy
from cognite.powerops.utils.powerops_status_events import create_bootstrap_finished_event
from cognite.powerops.utils.resource_generation import (
    generate_relationships_from_price_area_to_price,
    generate_resources_and_data,
)


def validate_config(
    config: BootstrapConfig,
    errors: Literal["keep", "fix"] = "keep",
) -> BootstrapConfig:
    config.validate_bid_configs()
    return config


def create_watercourse_timeseries_mappings(
    watercourse_configs: list[WatercourseConfig],
    time_series_mappings: List[TimeSeriesMapping],
) -> BootstrapResourceCollection:
    cdf_resources = BootstrapResourceCollection()
    for watercourse_config, time_series_mapping in zip(watercourse_configs, time_series_mappings):
        cdf_resources += create_base_mapping_bootstrap_resources(
            watercourse_config=watercourse_config,
            time_series_mapping=time_series_mapping,
        )
    return cdf_resources


def create_watercourse_processed_shop_files(
    watercourse_configs: list[WatercourseConfig],
) -> BootstrapResourceCollection:
    cdf_resources = BootstrapResourceCollection()
    for watercourse_config in watercourse_configs:
        process_yaml_file(
            yaml_raw_path=watercourse_config.yaml_raw_path,
            yaml_processed_path=watercourse_config.yaml_processed_path,
        )

        cdf_resources.add(
            [
                ShopFileConfig(
                    path=watercourse_config.yaml_processed_path,
                    cogshop_file_type="model",
                    watercourse_name=watercourse_config.name,
                )
            ]
        )

        # Create ShopOutputConfig Sequence
        shop_output_config = ShopOutputConfig(watercourse=watercourse_config.name)
        cdf_resources += shop_output_config.to_bootstrap_resources()

    return cdf_resources


def create_watercourse_shop_files(
    shop_file_configs: list[ShopFileConfig], watercourse_directories: dict
) -> list[ShopFileConfig]:
    for shop_file in shop_file_configs:
        shop_file.set_full_path(watercourse_directories[shop_file.watercourse_name])
    return shop_file_configs


def create_base_mapping_bootstrap_resources(
    watercourse_config: WatercourseConfig,
    time_series_mapping: TimeSeriesMapping,
) -> BootstrapResourceCollection:
    return write_mapping_to_sequence(
        mapping=time_series_mapping,
        watercourse=watercourse_config.name,
        mapping_type="base_mapping",
    )


def process_bid_process_configs(
    path: Path,
    bid_process_configs: list[BidProcessConfig],
    bidmatrix_generators: list[BidMatrixGeneratorConfig],
    price_scenarios_by_id: dict[str, PriceScenario],
    watercourses: list[WatercourseConfig],
    benchmark: BenchmarkingConfig,
    existing_bootstrap_resources: BootstrapResourceCollection,
) -> BootstrapResourceCollection:
    new_resources = BootstrapResourceCollection()
    for bid_process_config in bid_process_configs:
        new_resources += bid_process_config.to_bootstrap_resources(
            path=path,
            bootstrap_resources=existing_bootstrap_resources,
            price_scenarios_by_id=price_scenarios_by_id,
            bid_matrix_generator_configs=bidmatrix_generators,
            watercourses=watercourses,
            benchmark=benchmark,
        )

    return new_resources


def process_rkom_bid_configs(
    rkom_bid_combination_configs: Optional[List[RKOMBidCombinationConfig]],
    rkom_market_config: Optional[RkomMarketConfig],
    rkom_bid_process: List[RKOMBidProcessConfig],
    price_scenarios_by_id: dict[str, PriceScenario],
    market_name: str,
) -> BootstrapResourceCollection:
    bootstrap_resource_collection = BootstrapResourceCollection()

    if not rkom_market_config:
        rkom_market_config = RkomMarketConfig.default()
    bootstrap_resource_collection.add(rkom_market_config.cdf_asset)

    for rkom_process_config in rkom_bid_process:
        bootstrap_resource_collection += rkom_process_config.to_bootstrap_resources(
            price_scenarios_by_id=price_scenarios_by_id, market_name=market_name
        )

    if rkom_bid_combination_configs:
        for rkom_bid_combination_config in rkom_bid_combination_configs:
            bootstrap_resource_collection.add(rkom_bid_combination_config.cdf_asset)

    return bootstrap_resource_collection


def get_shop_service_url(cognite_project: str):
    return (
        "https://shop-production.az-inso-powerops.cognite.ai/submit-run"
        if cognite_project.endswith("-prod")
        else "https://shop-staging.az-inso-powerops.cognite.ai/submit-run"
    )


def _load_config(path: Path) -> BootstrapConfig:
    config = BootstrapConfig.from_yamls(path)
    if os.environ.get("COGNITE_PROJECT"):
        config.cdf = config.cdf.from_env()
    return config


def _transform(
    config: BootstrapConfig,
    path: Path,
    market: str = "Dayahead",
) -> BootstrapResourceCollection:
    cognite_project = config.cdf.dict()["COGNITE_PROJECT"]

    constants = config.constants
    _ = [w.set_shop_yaml_paths(path) for w in config.watercourses]
    watercourse_directories = {w.name: "/".join((path / w.directory).parts) for w in config.watercourses}

    shop_files_config_list = ShopFileConfigs.from_yaml(path).watercourses_shop

    print(f"Running bootstrap for data set {constants.data_set_external_id} in CDF project" + f"{cognite_project}")

    shop_service_url = get_shop_service_url(cognite_project)

    # TODO: split code below to three main functions(?) when having figured out function signature

    bootstrap_resources = BootstrapResourceCollection()
    # Create common CDF resources
    bootstrap_resources.add(create_labels())
    bootstrap_resources.add(
        create_skeleton_asset_hierarchy(
            organization_subdomain=constants.organization_subdomain,
            tenant_id=constants.tenant_id,
            shop_service_url=shop_service_url,
        )
    )
    # PowerOps asset data model
    bootstrap_resources += generate_resources_and_data(
        watercourse_configs=config.watercourses,
        plant_time_series_mappings=config.plant_time_series_mappings,
        generator_time_series_mappings=config.generator_time_series_mappings,
    )

    bootstrap_resources.add(generate_relationships_from_price_area_to_price(config.dayahead_price_timeseries))
    # SHOP files (model, commands, cut mapping++) and configs (base mapping, output definition)

    # Shop files related to each watercourse
    bootstrap_resources.add(create_watercourse_shop_files(shop_files_config_list, watercourse_directories))
    bootstrap_resources += create_watercourse_processed_shop_files(watercourse_configs=config.watercourses)
    bootstrap_resources += create_watercourse_timeseries_mappings(
        watercourse_configs=config.watercourses, time_series_mappings=config.time_series_mappings
    )

    # PowerOps configuration resources
    bootstrap_resources.add(config.market.cdf_asset)
    benchmarking_config_assets = [config.cdf_asset for config in config.benchmarks]
    bootstrap_resources.add(benchmarking_config_assets)

    bootstrap_resources += process_bid_process_configs(
        path=path,
        bid_process_configs=config.bidprocess,
        bidmatrix_generators=config.bidmatrix_generators,
        price_scenarios_by_id=config.price_scenario_by_id,
        watercourses=config.watercourses,
        benchmark=config.benchmarks[0],
        existing_bootstrap_resources=bootstrap_resources.copy(),
    )

    bootstrap_resources += process_rkom_bid_configs(
        rkom_bid_combination_configs=config.rkom_bid_combination,
        rkom_market_config=config.rkom_market,
        rkom_bid_process=config.rkom_bid_process,
        price_scenarios_by_id=config.price_scenario_by_id,
        market_name=market,
    )

    # Set hashes for Shop Files, needed for comparison
    for shop_config in bootstrap_resources.shop_file_configs.values():
        if shop_config.md5_hash is None:
            file_content = Path(shop_config.path).read_bytes()  # type: ignore[arg-type]
            shop_config.set_md5_hash(file_content)

    # ! This should always stay at the bottom # TODO: consider wrapper
    bootstrap_resources.add(create_bootstrap_finished_event())

    return bootstrap_resources


def _preview_resources_diff(
    bootstrap_resources: BootstrapResourceCollection,
    cdf_parameters: dict,
    data_set_external_id: str,
) -> None:
    # Preview differences between bootstrap resources and CDF resources
    client = get_client(cdf_parameters)
    cdf_bootstrap_resources = BootstrapResourceCollection.from_cdf(
        client=client, data_set_external_id=data_set_external_id
    )

    print(BootstrapResourceCollection.prettify_differences(bootstrap_resources.difference(cdf_bootstrap_resources)))


def _create_cdf_resources(
    bootstrap_resources: BootstrapResourceCollection,
    cdf_parameters: dict,
    data_set_external_id: str,
    overwrite_data: bool,
    skip_dm: bool,
) -> None:
    client = get_client(cdf_parameters)
    bootstrap_resources.write_to_cdf(
        client=client,
        data_set_external_id=data_set_external_id,
        overwrite=overwrite_data,
        skip_dm=skip_dm,
    )



from __future__ import annotations

import json
from dataclasses import dataclass
from functools import cached_property
from typing import ClassVar, Optional

from cognite.client.data_classes import Event

from cognite.powerops.utils.common import unique_str


@dataclass
class ShopRunEvent:
    """
    A cut-down variant of `common.workflow_utils.ShopRun` from
    power-ops-functions repo. This variant has no workflow event and
    no mappings.
    """

    event_type: ClassVar[str] = "POWEROPS_PROCESS_REQUESTED"
    process_type: ClassVar[str] = "POWEROPS_SHOP_RUN"
    watercourse: str
    starttime: str
    endtime: str
    timeresolution: Optional[dict[str, int]] = None
    dynamic_minute_offset: Optional[int] = None
    dm_case: Optional[str] = None
    dm_space: Optional[str] = None
    manual_run: bool = False

    def __post_init__(self):
        if self.starttime:
            self.starttime = str(self.starttime)
        if self.endtime:
            self.endtime = str(self.endtime)
        if self.timeresolution:
            self.timeresolution = {str(k): v for k, v in self.timeresolution.items()}

    @cached_property
    def external_id(self) -> str:
        return f"{self.process_type}_{unique_str()}"

    @property
    def metadata(self) -> dict:
        specific_metadata = {
            "shop:watercourse": self.watercourse,
            "shop:starttime": self.starttime,
            "shop:endtime": self.endtime,
            "process_type": self.process_type,
            "shop:manual_run": self.manual_run,
        }
        if self.dm_case is not None:
            specific_metadata["dm:case"] = self.dm_case
        if self.dm_space is not None:
            specific_metadata["dm:space"] = self.dm_space
        if self.timeresolution is not None:
            specific_metadata["shop:timeresolution"] = json.dumps(self.timeresolution)
        if self.dynamic_minute_offset is not None:
            specific_metadata["shop:dynamic_minute_offset"] = str(self.dynamic_minute_offset)
        return specific_metadata

    def to_dict(self, dataset_id: int) -> dict:
        return {
            "external_id": self.external_id,
            "data_set_id": dataset_id,
            "type": self.event_type,
            "metadata": self.metadata,
        }

    def to_event(self: ShopRunEvent, dataset_id: int) -> Event:
        return Event(**self.to_dict(dataset_id))

    @classmethod
    def from_event(cls, event: Event) -> ShopRunEvent:
        instance = ShopRunEvent(
            watercourse=event.metadata["shop:watercourse"],
            starttime=event.metadata["shop:starttime"],
            endtime=event.metadata["shop:endtime"],
            timeresolution=json.loads(event.metadata.get("shop:timeresolution", "null")),
            dynamic_minute_offset=event.metadata.get("shop:dynamic_minute_offset"),
            manual_run=event.metadata.get("manual_run", False),
        )
        if event.external_id:
            instance.external_id = event.external_id
        return instance



"""
This file is auto-generated using `dm topython` CLI tool.
"""

from __future__ import annotations

from typing import Optional

from cachelib import BaseCache, SimpleCache
from cognite.client import ClientConfig
from cognite.dm_clients.cdf.get_client import get_client_config
from cognite.dm_clients.config import settings
from cognite.dm_clients.domain_modeling import DomainClient, DomainModelAPI

from .schema import (
    Case,
    CommandsConfig,
    FileRef,
    Mapping,
    ModelTemplate,
    ProcessingLog,
    Scenario,
    Transformation,
    power_ops_dm_schema,
)


class PowerOpsDmClient(DomainClient):
    """Domain-specific client class for the 'PowerOpsDm'."""

    case: DomainModelAPI[Case]
    commands_config: DomainModelAPI[CommandsConfig]
    file_ref: DomainModelAPI[FileRef]
    mapping: DomainModelAPI[Mapping]
    model_template: DomainModelAPI[ModelTemplate]
    processing_log: DomainModelAPI[ProcessingLog]
    scenario: DomainModelAPI[Scenario]
    transformation: DomainModelAPI[Transformation]


def get_power_ops_dm_client(
    cache: Optional[BaseCache] = None,
    space_id: Optional[str] = None,
    data_model: Optional[str] = None,
    schema_version: Optional[int] = None,
    config: ClientConfig = None,
) -> PowerOpsDmClient:
    """Quick way of instantiating a PowerOpsDmClient with sensible defaults for development."""
    cache = cache if cache is not None else SimpleCache()
    space_external_id = space_id or settings.dm_clients.space
    data_model = data_model or settings.dm_clients.datamodel
    schema_version = schema_version or settings.dm_clients.schema_version
    config = config or get_client_config()
    return PowerOpsDmClient(
        power_ops_dm_schema,
        DomainModelAPI,
        cache,
        config,
        space_external_id,
        data_model,
        schema_version,
    )



from __future__ import annotations

import abc
import logging
import operator
import os
from datetime import datetime
from functools import reduce
from pathlib import Path
from typing import Generic, Sequence, TypeVar, Union

import yaml
from cognite.client.data_classes import FileMetadata
from matplotlib import pyplot as plt

from cognite.powerops.utils.plotting import ax_plot_time_series, create_time_series_plot

logger = logging.getLogger(__name__)


FileContentTypeT = TypeVar("FileContentTypeT", bound=Union[str, dict])


class ShopResultFile(abc.ABC, Generic[FileContentTypeT]):
    """Base class for handling a results file from Shop."""

    def __init__(self, content: FileContentTypeT, file_metadata: FileMetadata = None, encoding="utf-8") -> None:
        self._file_metadata = file_metadata
        self._encoding = encoding
        super().__init__()
        self.data: FileContentTypeT = content

    @property
    def encoding(self) -> str:
        return self._encoding

    @property
    def external_id(self):
        return self._file_metadata.external_id

    @property
    def name(self):
        return self._file_metadata.name

    def save_to_disk(self, dir_path: str = "") -> str:
        """Save file to local filesystem."""
        file_path = os.path.join(dir_path or os.getcwd(), self._file_metadata.external_id)
        Path(dir_path).mkdir(parents=True, exist_ok=True)
        with open(file_path, "w", encoding=self.encoding) as out_file:
            out_file.write(self.file_content)
        return file_path

    @property
    def file_content(self) -> str:
        """Data encoded for writing to local filesystem."""
        raise NotImplementedError()


class ShopLogFile(ShopResultFile[str]):
    """Plain text result file (for SHOP messages and CPlex logs)."""

    @property
    def file_content(self) -> str:
        return self.data


class ShopYamlFile(ShopResultFile[dict]):
    """Yaml-formatted results file (for post_run.yaml file created by SHOP)."""

    @property
    def file_content(self) -> str:
        return yaml.safe_dump(self.data, sort_keys=False)

    def _retrieve_time_series_dict(self, key: str) -> dict[datetime, float]:
        # key is a dot separated string of nested keys
        try:
            # FIXME? This is a workaround for the nested keys without dot get and
            # the fact that the yaml parser parses numeric keys as numbers
            keys = [int(k) if k.isdigit() else k for k in key.split(".")]
            data = reduce(operator.getitem, keys, self.data)
            if not (
                isinstance(data, dict)
                and all(isinstance(k, datetime) for k in data.keys())
                and all(isinstance(v, (float, int)) for v in data.values())
            ):
                raise ValueError
            return data
        except KeyError:
            logger.error(f'Key "{key}" not found in {self.name}')
        except ValueError:
            logger.error(f'Data at with key "{key}" cannot be plotted as a time series')
        return {}

    def _prepare_plot_time_series(self, keys: Union[str, Sequence[str]]) -> dict:
        if isinstance(keys, str):
            keys = [keys]
        return {key: self._retrieve_time_series_dict(key) for key in keys}

    def _case_insensitive_filter_out(self, str_list: list[str], to_match: Union[str, int]) -> bool:
        """Some keys are parsed as numbers by the yaml parser"""
        return all(str(to_match).lower() != str_in_list.lower() for str_in_list in str_list)

    def find_time_series(
        self,
        matches_object_types: Sequence[str] | str = "",
        matches_object_names: Sequence[str] | str = "",
        matches_attribute_names: Sequence[str] | str = "",
    ) -> list[str]:
        """Find time series in the results file"""
        if matches_object_types and isinstance(matches_object_types, str):
            matches_object_types = [matches_object_types]

        if matches_object_names and isinstance(matches_object_names, str):
            matches_object_names = [matches_object_names]

        if matches_attribute_names and isinstance(matches_attribute_names, str):
            matches_attribute_names = [matches_attribute_names]

        keys = []
        model = self.data["model"]
        for key1 in model:
            if matches_object_types and self._case_insensitive_filter_out(matches_object_types, key1):
                continue
            object_type: dict[str, dict] = model[key1]
            for key2, object_name in object_type.items():
                if matches_object_names and self._case_insensitive_filter_out(
                    matches_object_names,
                    key2,
                ):
                    continue
                for key3 in object_name:
                    if matches_attribute_names and self._case_insensitive_filter_out(
                        matches_attribute_names,
                        key3,
                    ):
                        continue
                    attribute = object_name[key3]
                    if isinstance(attribute, dict) and all(isinstance(x, datetime) for x in attribute.keys()):
                        keys.append(f"model.{key1}.{key2}.{key3}")
        return keys

    def plot(self, keys=Union[str, Sequence[str]]):
        if time_series := self._prepare_plot_time_series(keys):
            ax = create_time_series_plot()
            for key, ts_data in time_series.items():
                label = " ".join(key.split(".")[1:]).capitalize()
                ax_plot_time_series(ax, ts_data, label)

            ax.legend()
            plt.show()



from cognite.client import CogniteClient

from cognite.powerops.data_classes.time_series_mapping import TimeSeriesMapping
from cognite.powerops.utils.common import print_warning


def filter_time_series_mappings(mapping: TimeSeriesMapping, client: CogniteClient) -> TimeSeriesMapping:
    """Filters out mapping entries where the expected CDF time series is not found."""
    if not mapping.rows:
        print_warning("TimeSeriesMapping is empty! No filtering possible!")
        return mapping

    external_ids_in_mapping = {
        mapping.time_series_external_id for mapping in mapping if mapping.time_series_external_id is not None
    }

    external_ids_in_cdf = {
        time_series.external_id
        for time_series in client.time_series.retrieve_multiple(
            external_ids=list(external_ids_in_mapping), ignore_unknown_ids=True
        )
    }

    if missing_from_cdf := external_ids_in_mapping - external_ids_in_cdf:
        print_warning(f"Time series found in mapping but missing from CDF: {missing_from_cdf}")

    return TimeSeriesMapping(
        rows=[
            entry
            for entry in mapping
            if entry.time_series_external_id in external_ids_in_cdf or entry.time_series_external_id is None
        ]
    )


def merge_and_keep_last_mapping_if_overlap(*mappings: TimeSeriesMapping) -> TimeSeriesMapping:
    """Within each TimeSeriesMapping keep the last TimeSeriesMappingEntry in cases
    where they have the same `shop_model_path`. Across mappings keep the last TimeSeriesMappingEntry
    from the last TimeSeriesMapping.
    """

    keep_last_across_lists = {entry.shop_model_path: entry for mapping in mappings for entry in mapping}

    return TimeSeriesMapping(rows=list(keep_last_across_lists.values()))



from __future__ import annotations

import logging
from typing import Union

import pandas as pd
from cognite.client import CogniteClient
from cognite.client.data_classes import Asset, Event, LabelDefinition, Relationship, Sequence
from cognite.client.data_classes._base import CogniteResource, CogniteResourceList
from deepdiff import DeepDiff
from pydantic import BaseModel, Extra

from cognite.powerops.data_classes.shop_file_config import ShopFileConfig
from cognite.powerops.utils.cdf_utils import upsert_cognite_resources
from cognite.powerops.utils.files import upload_shop_config_file

logger = logging.getLogger(__name__)


ExternalId = str

AddableResourceT = Union["CogniteResource", list["CogniteResource"], "SequenceContent", list[ShopFileConfig]]


class SequenceRows(BaseModel):
    rows: list[tuple[int, list[str]]]
    columns_external_ids: list[str]

    def to_pandas(self) -> pd.DataFrame:
        return pd.DataFrame(
            data=[row[1] for row in self.rows],
            index=[row[0] for row in self.rows],
            columns=self.columns_external_ids,
        )


class SequenceContent(BaseModel):
    sequence_external_id: str
    data: Union[pd.DataFrame, SequenceRows]

    class Config:
        arbitrary_types_allowed = True

    def dump(self) -> dict:
        return {
            self.sequence_external_id: self.data.to_dict()
            if isinstance(self.data, pd.DataFrame)
            else self.data.to_pandas().to_dict()
        }


def clean_cdf_resources_for_diff(cdf_resources: dict[ExternalId, dict]) -> None:
    """Remove fields that are not relevant for diffing"""
    dynamic_fields = ["last_updated_time", "created_time", "parent_id", "root_id", "data_set_id", "id"]
    for resource in cdf_resources.values():
        for field in dynamic_fields:
            resource.pop(field, None)
        # remove the metadata if it is empty
        metadata = resource.get("metadata", {})
        if metadata == {}:
            resource.pop("metadata", None)
        else:
            for key, val in metadata.items():
                if isinstance(val, float):
                    # convert float to str
                    metadata[key] = str(val)

        # For sequences remove the column fields createdTime, lastUpdatedTime and metadata for each column
        if resource.get("columns"):
            for column in resource["columns"]:
                column.pop("createdTime", None)
                column.pop("lastUpdatedTime", None)
                column.pop("metadata", None)


def clean_local_resources_for_diff(local_resources: dict[ExternalId, dict]) -> None:
    for resource in local_resources.values():
        metadata = resource.get("metadata", {})
        for key, val in metadata.items():
            if isinstance(val, (float, int)):
                # convert float to str
                metadata[key] = str(val)


class BootstrapResourceCollection(BaseModel):
    # CDF resources
    assets: dict[ExternalId, Asset] = {}
    relationships: dict[ExternalId, Relationship] = {}
    sequences: dict[ExternalId, Sequence] = {}
    label_definitions: dict[ExternalId, LabelDefinition] = {}
    events: dict[ExternalId, Event] = {}

    # other resources
    sequence_content: dict[ExternalId, SequenceContent] = {}
    shop_file_configs: dict[ExternalId, ShopFileConfig] = {}

    class Config:
        arbitrary_types_allowed = True
        extra = Extra.forbid

    @property
    def all_cdf_resources(self) -> list[CogniteResource]:
        return (
            list(self.assets.values())
            + list(self.relationships.values())
            + list(self.sequences.values())
            + list(self.label_definitions.values())
            + list(self.events.values())
        )

    def add(self, resources_to_append: AddableResourceT):
        # sort the resource by type and append to the correct list
        if isinstance(resources_to_append, (list, CogniteResourceList)):
            for resource in resources_to_append:
                self._add_resource(resource)
        else:
            self._add_resource(resources_to_append)

    def _add_resource(self, resource: CogniteResource | SequenceContent | ShopFileConfig):
        if isinstance(resource, Asset):
            self.assets[resource.external_id] = resource
        elif isinstance(resource, Sequence):
            self.sequences[resource.external_id] = resource
        elif isinstance(resource, Relationship):
            self.relationships[resource.external_id] = resource
        elif isinstance(resource, LabelDefinition):
            self.label_definitions[resource.external_id] = resource
        elif isinstance(resource, Event):
            self.events[resource.external_id] = resource
        elif isinstance(resource, SequenceContent):
            self.sequence_content[resource.sequence_external_id] = resource
        elif isinstance(resource, ShopFileConfig):
            self.shop_file_configs[resource.external_id] = resource
        else:
            raise ValueError(f"Unknown resource type: {type(resource)}")

    def write_to_cdf(
        self,
        client: CogniteClient,
        data_set_external_id: str,
        overwrite: bool = False,
        skip_dm: bool = False,
    ):
        # CDF auth and data set
        data_set_id = client.data_sets.retrieve(external_id=data_set_external_id).id
        for resource in self.all_cdf_resources:
            resource.data_set_id = data_set_id

        api_by_type = {
            "label_definitions": client.labels,
            "assets": client.assets,
            "sequences": client.sequences,
            "relationships": client.relationships,
            "events": client.events,
        }
        for resource_type, api in api_by_type.items():
            resources = list(getattr(self, resource_type).values())
            logger.debug(f"Processing {len(resources)} {resource_type}...")
            upsert_cognite_resources(api, resource_type, resources)

        logger.debug(f"Processing {len(self.sequence_content)} sequences...")
        for sequence_external_id, sequence_data in self.sequence_content.items():
            if isinstance(sequence_data.data, pd.DataFrame):
                try:
                    client.sequences.data.insert_dataframe(
                        dataframe=sequence_data.data, external_id=sequence_external_id
                    )
                except Exception as e:
                    logger.warning(f"Failed to insert sequence dataframe for {sequence_external_id}: {e}")
            else:
                try:
                    client.sequences.data.insert(
                        rows=sequence_data.data.rows,
                        external_id=sequence_external_id,
                        column_external_ids=sequence_data.data.columns_external_ids,
                    )
                except Exception as e:
                    logger.warning(f"Failed to insert sequence rows for {sequence_external_id}: {e}")

        logger.debug(f"Processing {len(self.shop_file_configs)} files...")
        for shop_config in self.shop_file_configs.values():
            upload_shop_config_file(client=client, config=shop_config, overwrite=overwrite, data_set_id=data_set_id)

    def __add__(self, other: "BootstrapResourceCollection") -> "BootstrapResourceCollection":
        return BootstrapResourceCollection(
            assets=self.assets | other.assets,
            relationships=self.relationships | other.relationships,
            label_definitions=self.label_definitions | other.label_definitions,
            events=self.events | other.events,
            sequences=self.sequences | other.sequences,
            sequence_content=self.sequence_content | other.sequence_content,
            shop_file_configs=self.shop_file_configs | other.shop_file_configs,
        )

    def __iadd__(self, other: "BootstrapResourceCollection") -> "BootstrapResourceCollection":
        self.assets |= other.assets
        self.relationships |= other.relationships
        self.label_definitions |= other.label_definitions
        self.events |= other.events
        self.sequences |= other.sequences
        self.sequence_content |= other.sequence_content
        self.shop_file_configs |= other.shop_file_configs
        return self

    def difference(self, cdf: "BootstrapResourceCollection") -> dict[str, str]:
        local = self

        resource_diff: dict[str, str] = {}

        # First, we handle the cdf resources (everything except sequence data and shop files)
        for cdf_resource in [
            "assets",
            "relationships",
            "sequences",
            "label_definitions",
            "events",
        ]:
            local_resources = {
                resource.external_id: resource.dump()
                for resource in getattr(local, cdf_resource).values()
                if not (isinstance(resource, Event) and resource.type == "POWEROPS_BOOTSTRAP_FINISHED")
            }
            clean_local_resources_for_diff(local_resources)
            cdf_resources = {
                resource.external_id: resource.dump()
                for resource in getattr(cdf, cdf_resource).values()
                if not (isinstance(resource, Event) and resource.type == "POWEROPS_BOOTSTRAP_FINISHED")
            }
            clean_cdf_resources_for_diff(cdf_resources)

            resource_diff[cdf_resource] = DeepDiff(cdf_resources, local_resources, ignore_string_case=True).pretty()

        # Then we handle the sequence content
        local_sequence_content = {external_id: data.dump() for external_id, data in local.sequence_content.items()}
        cdf_sequence_content = {external_id: data.dump() for external_id, data in cdf.sequence_content.items()}
        sequence_content_diff = {
            "sequence_content": DeepDiff(
                cdf_sequence_content, local_sequence_content, ignore_nan_inequality=True
            ).pretty()
        }

        # And then finally the shop files
        shop_file_diff = {
            "shop_files": DeepDiff(
                {external_id: config.dict(exclude={"path"}) for external_id, config in cdf.shop_file_configs.items()},
                {external_id: config.dict(exclude={"path"}) for external_id, config in local.shop_file_configs.items()},
            ).pretty()
        }

        return resource_diff | sequence_content_diff | shop_file_diff

    @classmethod
    def prettify_differences(cls, diff_per_resource_type: dict[str, str]) -> str:
        output_message = ""
        for resource_type, differences in diff_per_resource_type.items():
            if differences:
                output_message += "--------------------------------------------\n"
                output_message += f"Difference for resource type {resource_type}: \n"
                output_message += "--------------------------------------------\n"
                output_message += differences
                output_message += "\n"
        return output_message.replace("added to dictionary.", "missing in CDF.").replace(
            "removed from dictionary.", "missing locally."
        )

    @classmethod
    def from_cdf(
        cls,
        client: CogniteClient,
        data_set_external_id: str,
    ) -> "BootstrapResourceCollection":
        """
        Function that creates a BootstrapResourceCollection from a CDF data set (typically the bootstrap data set)
        """

        data_set_id = client.data_sets.retrieve(external_id=data_set_external_id).id
        bootstrap_resource_collection = BootstrapResourceCollection()

        # start by downloading all the resources from CDF
        for resource_type in [
            "assets",
            "sequences",
            "relationships",
            "labels",
            "events",
        ]:
            # get the api for the resource type
            api = getattr(client, resource_type)
            # get all the resources from CDF
            resources = api.list(data_set_ids=[data_set_id], limit=None)
            # add the resources to the bootstrap resource collection
            bootstrap_resource_collection.add(resources)

        file_meta = client.files.list(data_set_ids=[data_set_id], limit=None)
        existing_external_ids = {f.external_id for f in file_meta}
        shop_files = []
        for f in file_meta:
            shop_config = ShopFileConfig.from_file_meta(f)
            if shop_config.md5_hash is None and shop_config.external_id in existing_external_ids:
                file_content = client.files.download_bytes(external_id=shop_config.external_id)
                shop_config.set_md5_hash(file_content)
            shop_files.append(shop_config)
        bootstrap_resource_collection.add(shop_files)

        # then download the sequence data
        for external_id in bootstrap_resource_collection.sequences:
            sequence_data = client.sequences.data.retrieve_dataframe(
                external_id=external_id, limit=None, start=0, end=-1
            )
            bootstrap_resource_collection.add(SequenceContent(sequence_external_id=external_id, data=sequence_data))

        return bootstrap_resource_collection



from __future__ import annotations

import ast
import json
import os
import typing
from collections import defaultdict
from pathlib import Path
from typing import ClassVar, Dict, Generator, List, Optional, Tuple

from cognite.client.data_classes import Asset, Label, Sequence
from pydantic import BaseModel, Field, root_validator, validator

from cognite.powerops.data_classes.cdf_resource_collection import (
    BootstrapResourceCollection,
    SequenceContent,
    SequenceRows,
)
from cognite.powerops.data_classes.common import AggregationMethod, CommonConstants, RelativeTime, RetrievalType
from cognite.powerops.data_classes.config_model import Configuration
from cognite.powerops.data_classes.reserve_scenario import Auction, Block, Product, ReserveScenario
from cognite.powerops.data_classes.time_series_mapping import (
    TimeSeriesMapping,
    TimeSeriesMappingEntry,
    write_mapping_to_sequence,
)
from cognite.powerops.data_classes.transformation import Transformation, TransformationType
from cognite.powerops.utils.cdf_utils import simple_relationship
from cognite.powerops.utils.common import print_warning
from cognite.powerops.utils.labels import AssetLabels, RelationshipLabels
from cognite.powerops.utils.serializer import load_yaml


class Watercourse(BaseModel):
    name: str
    shop_penalty_limit: int = 42000


class WatercourseConfig(Watercourse):
    """
    Represents the configuration for a Watercourse

    Attributes:
        version: The version of the watercourse configuration.

    """

    version: str
    market_to_price_area: Dict[str, str]
    directory: str
    model_raw: str
    # ------------------------------------------------------------------
    yaml_raw_path: str = ""
    yaml_processed_path: str = ""  # TODO: not used here
    yaml_mapping_path: str = ""
    model_processed: str  # TODO: not used here
    model_mapping: Optional[str] = None
    tco_paths: Optional[List[str]] = None  # TODO: not used here - HEV specific
    rrs_ids: Optional[List[str]] = None
    hardcoded_mapping: Optional[TimeSeriesMapping] = None  # TODO: not used here
    hist_flow_timeseries: Optional[Dict[str, str]] = None  # TODO: not used here
    # ------------------------------------------------------------------
    production_obligation_ts_ext_ids: Optional[List[str]] = None
    plant_display_names_and_order: Optional[Dict[str, tuple[str, int]]] = None
    reservoir_display_names_and_order: Optional[Dict[str, tuple[str, int]]] = None
    water_value_based_method_time_series_csv_filename: Optional[str] = None

    @classmethod
    def from_yaml(cls, yaml_path: Path) -> list["WatercourseConfig"]:
        return [cls(**watercourse_raw) for watercourse_raw in load_yaml(yaml_path)]

    def load_plants_by_price_area(self, path: Path) -> dict[str, list]:
        content = load_yaml(path / self.directory / self.model_raw, clean_data=True)
        plants_by_prod_area = defaultdict(list)
        for plant_name, plant_attributes in content["model"]["plant"].items():
            prod_area = str(list(plant_attributes["prod_area"].values())[0])
            price_area = self.market_to_price_area[prod_area]
            plants_by_prod_area[price_area].append(plant_name)
        return plants_by_prod_area

    def plant_display_name(self, plant: str) -> Optional[str]:
        try:
            return self.plant_display_names_and_order[plant][0]  # type: ignore
        except (KeyError, TypeError):
            print(f"[WARNING] No display name for plant: {plant}")
            return None

    def plant_ordering_key(self, plant: str) -> Optional[int]:
        try:
            return self.plant_display_names_and_order[plant][1]  # type: ignore
        except (KeyError, TypeError):
            print(f"[WARNING] No ordering key for plant: {plant}")
            return None

    def reservoir_display_name(self, reservoir: str) -> Optional[str]:
        return self.reservoir_display_names_and_order[reservoir][0] if self.reservoir_display_names_and_order else None

    def reservoir_ordering_key(self, reservoir: str) -> Optional[int]:
        return self.reservoir_display_names_and_order[reservoir][1] if self.reservoir_display_names_and_order else None

    def set_shop_yaml_paths(self, path):
        self.yaml_raw_path = f"{path}/{self.directory}/{self.model_raw}"  # TODO: create these as properties
        self.yaml_processed_path = f"{path}/{self.directory}/{self.model_processed}"
        self.yaml_mapping_path = f"{path}/{self.directory}/{self.model_mapping}"


class PriceScenarioID(BaseModel):
    id: str
    rename: str = ""


class PriceScenario(BaseModel):
    name: str
    time_series_external_id: Optional[str]
    transformations: Optional[List[Transformation]]

    def to_time_series_mapping(self) -> TimeSeriesMapping:
        retrieve = RetrievalType.RANGE if self.time_series_external_id else None
        transformations = self.transformations or []

        # to make buy price slightly higher than sale price in SHOP
        transformations_buy_price = transformations + [
            Transformation(transformation=TransformationType.ADD, kwargs={"value": 0.01})
        ]

        sale_price_row = TimeSeriesMappingEntry(
            object_type="market",
            object_name=self.name,
            attribute_name="sale_price",
            time_series_external_id=self.time_series_external_id,
            transformations=transformations,
            retrieve=retrieve,
            aggregation=AggregationMethod.mean,
        )

        buy_price_row = TimeSeriesMappingEntry(
            object_type="market",
            object_name=self.name,
            attribute_name="buy_price",
            time_series_external_id=self.time_series_external_id,
            transformations=transformations_buy_price,
            retrieve=retrieve,
            aggregation=AggregationMethod.mean,
        )

        return TimeSeriesMapping(rows=[sale_price_row, buy_price_row])


class BidMatrixGeneratorConfig(BaseModel):
    name: str
    default_method: str
    default_function_external_id: str
    column_external_ids: ClassVar[list[str]] = [
        "shop_plant",
        "bid_matrix_generation_method",
        "function_external_id",
    ]

    @staticmethod
    def _metadata(price_area: str, bid_process_config_name: str) -> dict:
        # TODO: Rename this from "shop:type" to something without "shop"
        #   (but check if e.g. power-ops-functions uses this)
        return {
            "bid:price_area": f"price_area_{price_area}",
            "shop:type": "bid_matrix_generator_config",
            "bid:bid_process_configuration_name": bid_process_config_name,
        }

    def get_sequence(self, price_area: str, bid_process_config_name: str) -> Sequence:
        column_def = [{"valueType": "STRING", "externalId": external_id} for external_id in self.column_external_ids]
        return Sequence(
            external_id=f"POWEROPS_bid_matrix_generator_config_{bid_process_config_name}",
            name=f"POWEROPS bid matrix generator config {bid_process_config_name}",
            description="Configuration of bid matrix generation method to use for each plant in the price area",
            columns=column_def,
            metadata=self._metadata(price_area, bid_process_config_name),
        )

    def to_sequence_rows(self, plant_names: list[str]) -> SequenceRows:
        return SequenceRows(
            rows=[
                (idx, [plant, self.default_method, self.default_function_external_id])
                for idx, plant in enumerate(plant_names)
            ],
            columns_external_ids=self.column_external_ids,
        )

    def to_bootstrap_resources(
        self,
        bid_process_config_asset: Asset,
        price_area: str,
        bid_process_config_name: str,
        plant_names: list[str],
    ) -> BootstrapResourceCollection:
        bootstrap_resource_collection = BootstrapResourceCollection()

        sequence = self.get_sequence(price_area, bid_process_config_name)
        bootstrap_resource_collection.add(sequence)

        sequence_rows = self.to_sequence_rows(plant_names)
        bootstrap_resource_collection.add(
            SequenceContent(sequence_external_id=sequence.external_id, data=sequence_rows)
        )

        relationship = simple_relationship(
            source=bid_process_config_asset,
            target=sequence,
            label_external_id=RelationshipLabels.BID_MATRIX_GENERATOR_CONFIG_SEQUENCE,
        )
        bootstrap_resource_collection.add(relationship)

        return bootstrap_resource_collection


class RkomMarketConfig(BaseModel):
    external_id: str
    name: str
    timezone: str
    start_of_week: int
    parent_external_id: ClassVar[str] = "market_configurations"

    @property
    def metadata(self) -> dict:
        return {
            "timezone": self.timezone,
            "start_of_week": self.start_of_week,
        }

    @property
    def cdf_asset(self) -> Asset:
        return Asset(
            external_id=self.external_id,
            name=self.name,
            metadata=self.metadata,
            parent_external_id=self.parent_external_id,
            labels=["market"],
        )

    @staticmethod
    def default() -> "RkomMarketConfig":
        return RkomMarketConfig(
            external_id="market_configuration_statnett_rkom_weekly",
            name="RKOM weekly (Statnett)",
            timezone="Europe/Oslo",
            start_of_week=1,
        )


class RKOMBidCombinationConfig(Configuration):
    auction: Auction = Field(alias="bid_auction")
    name: str = Field("default", alias="bid_combination_name")
    rkom_bid_config_external_ids: List[str] = Field(alias="bid_rkom_bid_configs")
    parent_external_id: ClassVar[str] = "rkom_bid_combination_configurations"

    @validator("auction", pre=True)
    def to_enum(cls, value):
        return Auction[value] if isinstance(value, str) else value

    @validator("rkom_bid_config_external_ids", pre=True)
    def parse_string(cls, value):
        return [external_id for external_id in ast.literal_eval(value)] if isinstance(value, str) else value

    @property
    def cdf_asset(self) -> Asset:
        sequence_external_id = f"RKOM_bid_combination_configuration_{self.auction.value}_{self.name}"

        return Asset(
            name=sequence_external_id.replace("_", " "),
            description="Defining which RKOM bid methods should be combined (into the total bid form)",
            external_id=sequence_external_id,
            metadata={
                "bid:auction": self.auction.value,
                "bid:combination_name": self.name,
                "bid:rkom_bid_configs": json.dumps(self.rkom_bid_config_external_ids),
            },
            parent_external_id=self.parent_external_id,
        )


class BenchmarkingConfig(Configuration):
    bid_date: RelativeTime
    shop_start: RelativeTime = Field(alias="shop_starttime")
    shop_end: RelativeTime = Field(alias="shop_endtime")
    production_plan_time_series: Optional[Dict[str, List[str]]]
    market_config_external_id: str = Field(alias="bid_market_config_external_id")
    relevant_shop_objective_metrics: Dict[str, str] = {
        "grand_total": "Grand Total",
        "total": "Total",
        "sum_penalties": "Sum Penalties",
        "major_penalties": "Major Penalties",
        "minor_penalties": "Minor Penalties",
        "load_value": "Load Value",
        "market_sale_buy": "Market Sale Buy",
        "rsv_end_value_relative": "RSV End Value Relative",
        "startup_costs": "Startup Costs",
        "vow_in_transit": "Vow in Transit",
        "sum_feeding_fee": "Sum Feeding Fee",
        "rsv_tactical_penalty": "RSV Tactical Penalty",
        "rsv_end_value": "RSV End Value",
        "bypass_cost": "Bypass Cost",
        "gate_discharge_cost": "Gate Discharge Cost",
        "reserve_violation_penalty": "Reserve Violation Penalty",
        "load_penalty": "Load Penalty",
    }  # Pydantic handles mutable defaults such that this is OK:
    # https://stackoverflow.com/questions/63793662/how-to-give-a-pydantic-list-field-a-default-value/63808835#63808835
    # TODO: Consider adding relationships to bid process config
    #  assets (or remove the optional part that uses those relationships in power-ops-functions)

    @validator("shop_start", "shop_end", "bid_date", pre=True)
    def json_loads(cls, value):
        return {"operations": json.loads(value)} if isinstance(value, str) else value

    @property
    def metadata(self) -> dict:
        metadata = {
            "bid:date": str(self.bid_date),
            "shop:starttime": str(self.shop_start),
            "shop:endtime": str(self.shop_end),
            "bid:market_config_external_id": self.market_config_external_id,
            "benchmarking_metrics": json.dumps(self.relevant_shop_objective_metrics),
        }
        if self.production_plan_time_series:
            metadata["benchmark:production_plan_time_series"] = json.dumps(
                self.production_plan_time_series, ensure_ascii=False
            )  # ensure_ascii=False to treat Nordic letters properly
        return metadata

    @property
    def cdf_asset(self) -> Asset:
        return Asset(
            external_id="POWEROPS_dayahead_bidding_benchmarking_config",
            name="Benchmarking config DA",
            description="Configuration for benchmarking of day-ahead bidding",
            metadata=self.metadata,
            parent_external_id="benchmarking_configurations",
            labels=["dayahead_bidding_benchmarking_config"],
        )


class MarketConfig(Configuration):
    external_id: str
    name: str
    max_price: float = None
    min_price: float = None
    time_unit: str = None
    timezone: str
    tick_size: float = None
    trade_lot: float = None
    price_steps: int = None
    parent_external_id: ClassVar[str] = "market_configurations"
    price_unit: str = None

    @property
    def metadata(self) -> dict:
        return {
            "min_price": self.min_price,
            "max_price": self.max_price,
            "timezone": self.timezone,
            "time_unit": self.time_unit,
            "tick_size": self.tick_size,
            "trade_lot": self.trade_lot,
            "price_steps": self.price_steps,
            "price_unit": self.price_unit,
        }

    @property
    def cdf_asset(self) -> Asset:
        return Asset(
            external_id=self.external_id,
            name=self.name,
            metadata=self.metadata,
            parent_external_id=self.parent_external_id,
            labels=["market"],
        )


MARKET_CONFIG_NORDPOOL_DAYAHEAD = MarketConfig(
    external_id="market_configuration_nordpool_dayahead",
    name="Nord Pool Day-ahead",
    max_price=4000,
    min_price=-500,
    time_unit="1h",
    timezone="Europe/Oslo",
    tick_size=0.1,
    trade_lot=0.1,
    price_steps=200,
    price_unit="EUR/MWh",
)


class BidProcessConfig(Configuration):
    name: str
    price_area_name: str = Field(alias="bid_price_area")
    price_scenarios: List[PriceScenarioID] = Field(alias="bid_price_scenarios")
    main_scenario: str = Field(alias="bid_main_scenario")
    bid_date: Optional[RelativeTime] = None
    shop_start: Optional[RelativeTime] = Field(None, alias="shop_starttime")
    shop_end: Optional[RelativeTime] = Field(None, alias="shop_endtime")
    bid_matrix_generator: str = Field(alias="bid_bid_matrix_generator_config_external_id")
    price_scenarios_per_watercourse: Optional[Dict[str, typing.Set[str]]] = None
    is_default_config_for_price_area: bool = False

    @validator("shop_start", "shop_end", "bid_date", pre=True)
    def json_loads(cls, value):
        return {"operations": json.loads(value)} if isinstance(value, str) else value

    @validator("price_scenarios", pre=True)
    def literal_eval(cls, value):
        return [{"id": id_} for id_ in ast.literal_eval(value)] if isinstance(value, str) else value

    def to_cdf_asset(
        self,
        benchmark: BenchmarkingConfig,
        price_scenarios_by_id: dict[str, PriceScenario],
    ) -> Asset:
        def to_metadata(price_scenarios_by_id: dict[str, PriceScenario], benchmark: BenchmarkingConfig) -> dict:
            price_scenarios = map_price_scenarios_by_name(
                self.price_scenarios, price_scenarios_by_id, MARKET_BY_PRICE_AREA[self.price_area_name]
            )

            price_scenarios_string = json.dumps(
                {scenario_name: scenario.time_series_external_id for scenario_name, scenario in price_scenarios.items()}
            )
            return {
                "bid:bid_process_configuration_name": self.name,
                "bid:bid_matrix_generator_config_external_id": f"POWEROPS_bid_matrix_generator_config_{self.name}",
                "bid:market_config_external_id": benchmark.market_config_external_id,
                "bid:price_scenarios": price_scenarios_string,
                "bid:main_scenario": self.main_scenario,
                "bid:date": str(self.bid_date or benchmark.bid_date),
                "shop:starttime": str(self.shop_start or benchmark.shop_start),
                "shop:endtime": str(self.shop_end or benchmark.shop_end),
                "bid:price_area": f"price_area_{self.price_area_name}",
                "bid:is_default_config_for_price_area": self.is_default_config_for_price_area,
            }

        return Asset(
            external_id=f"POWEROPS_bid_process_configuration_{self.name}",
            name=f"Bid process configuration {self.name}",
            metadata=to_metadata(price_scenarios_by_id, benchmark),
            description="Bid process configuration defining how to run a bid matrix generation process",
            parent_external_id="bid_process_configurations",
            labels=[Label(AssetLabels.BID_PROCESS_CONFIGURATION)],
        )

    def get_price_scenarios(
        self,
        price_scenarios: dict[str, PriceScenario],  # TODO: This must be price_scenarios_per_watercourse
        watercourse_name: Optional[str] = None,
    ) -> Dict[str, PriceScenario]:
        if watercourse_name and self.price_scenarios_per_watercourse:
            try:
                return {
                    scenario_name: price_scenarios[scenario_name]
                    for scenario_name in self.price_scenarios_per_watercourse[watercourse_name]
                }
            except KeyError as e:
                raise KeyError(
                    f"Watercourse {watercourse_name} not defined in price_scenarios_per_watercourse "
                    f"for BidProcessConfig {self.name}"
                ) from e
        return price_scenarios

    def to_bootstrap_resources(
        self,
        path: Path,
        bootstrap_resources: BootstrapResourceCollection,
        price_scenarios_by_id: dict[str, PriceScenario],
        bid_matrix_generator_configs: list[BidMatrixGeneratorConfig],
        watercourses: list[WatercourseConfig],
        benchmark: BenchmarkingConfig,
    ) -> BootstrapResourceCollection:
        asset = self.to_cdf_asset(benchmark, price_scenarios_by_id)
        price_scenario_by_name = map_price_scenarios_by_name(
            self.price_scenarios, price_scenarios_by_id, MARKET_BY_PRICE_AREA[self.price_area_name]
        )
        bid_matrix_generator_config_resources = self.create_bid_matrix_generator_config_resources(
            bid_matrix_generator_configs,
            asset,
            path,
            watercourses,
        )

        watercourse_names = [
            bootstrap_resources.assets[rel.target_external_id].name
            for rel in bootstrap_resources.relationships.values()
            if rel.source_external_id == f"price_area_{self.price_area_name}"
            and RelationshipLabels.WATERCOURSE in rel.labels[0].values()
        ]

        if not watercourse_names:
            print_warning(
                f"No related watercourses found for price area {self.price_area_name}! No 'incremental mapping' "
                f"Sequences to write!"
            )
            return bid_matrix_generator_config_resources

        return bid_matrix_generator_config_resources + self.create_incremental_mapping_resources(
            price_scenario_by_name, watercourse_names, asset
        )

    def create_incremental_mapping_resources(
        self,
        price_scenario_by_name: dict[str, PriceScenario],
        watercourse_names: list[str],
        asset: Asset,
    ):
        # Create TimeSeriesMapping (incremental mapping) for each price scenario and watercourse
        incremental_mapping_bootstrap_resources = BootstrapResourceCollection()
        for watercourse in watercourse_names:
            for scenario_name, price_scenario in self.get_price_scenarios(price_scenario_by_name, watercourse).items():
                time_series_mapping = price_scenario.to_time_series_mapping()

                incremental_mapping_bootstrap_resources += write_mapping_to_sequence(
                    watercourse=watercourse,
                    mapping=time_series_mapping,
                    mapping_type="incremental_mapping",
                    price_scenario_name=scenario_name,
                    config_name=self.name,
                )

        for sequence in incremental_mapping_bootstrap_resources.sequences.values():
            relationship = simple_relationship(
                source=asset,
                target=sequence,
                label_external_id=RelationshipLabels.INCREMENTAL_MAPPING_SEQUENCE,
            )
            incremental_mapping_bootstrap_resources.add(relationship)
        return incremental_mapping_bootstrap_resources

    def create_bid_matrix_generator_config_resources(
        self,
        bid_matrix_generator_configs: list[BidMatrixGeneratorConfig],
        asset: Asset,
        path: Path,
        watercourses: list[WatercourseConfig],
    ):
        new_bootstrap_resources = BootstrapResourceCollection()
        new_bootstrap_resources.add(asset)

        plants_by_price_area: dict[str, list] = defaultdict(list)
        for w in watercourses:
            sub_dict = w.load_plants_by_price_area(path)
            for k, v in sub_dict.items():
                plants_by_price_area[k].extend(v)

        generator_by_name = {g.name: g for g in bid_matrix_generator_configs}

        # Create BidMatrixGeneratorConfig Sequence and Relationship to BidProcessConfig Asset
        new_bootstrap_resources += generator_by_name[self.bid_matrix_generator].to_bootstrap_resources(
            bid_process_config_asset=asset,
            price_area=self.price_area_name,
            bid_process_config_name=self.name,
            plant_names=plants_by_price_area[self.price_area_name],
        )
        return new_bootstrap_resources


# TODO: Put this in a configgen module?
def map_price_scenarios_by_name(
    scenario_ids: List[PriceScenarioID], price_scenarios_by_id: dict[str, PriceScenario], market_name: str
) -> dict[str, PriceScenario]:
    scenario_by_name = {}
    for identifier in scenario_ids:
        ref_scenario = price_scenarios_by_id[identifier.id]
        name = identifier.rename or ref_scenario.name or identifier.id
        scenario_by_name[name] = PriceScenario(name=market_name, **ref_scenario.dict(exclude={"name"}))
    return scenario_by_name


MARKET_BY_PRICE_AREA = {"NO2": "Dayahead", "NO1": "1", "NO3": "1", "NO5": "1"}


class CDFConfig(BaseModel):
    TENANT_ID: str
    CLIENT_ID: str
    CDF_CLUSTER: str
    COGNITE_PROJECT: str
    SPACE_ID: str
    DATA_MODEL: str
    SCHEMA_VERSION: str

    @classmethod
    def from_env(cls) -> "CDFConfig":
        return cls(
            **{
                "TENANT_ID": os.environ.get("TENANT_ID"),
                "CLIENT_ID": os.environ.get("CLIENT_ID"),
                "CLIENT_SECRET_ENV": os.environ.get("CLIENT_SECRET_ENV"),
                "CDF_CLUSTER": os.environ.get("CDF_CLUSTER"),
                "COGNITE_PROJECT": os.environ.get("COGNITE_PROJECT"),
                "SPACE_ID": os.environ.get("SPACE_ID"),
                "DATA_MODEL": os.environ.get("DATA_MODEL"),
                "SCHEMA_VERSION": os.environ.get("SCHEMA_VERSION"),
            }
        )


class ReserveScenarios(BaseModel):
    volumes: list[int]
    auction: Auction
    product: Product
    block: Block
    reserve_group: str
    mip_plant_time_series: List[Tuple[str, Optional[str]]]
    obligation_external_id: Optional[str]

    @validator("auction", pre=True)
    def to_enum(cls, value):
        return Auction[value] if isinstance(value, str) else value

    @validator("volumes", pre=True)
    def valid_volumes(cls, volumes):
        if 0 not in volumes:
            raise ValueError("You probably want 0 MW as one of the volumes!")
        if any(volume < 0 for volume in volumes):
            raise ValueError(f"All volumes should be positive! Got {volumes}")
        return list(set(volumes))  # Do not want duplicate volumes

    def __str__(self) -> str:
        return f'[{"MW, ".join(str(volume) for volume in sorted(self.volumes))}MW]'  # E.g "[0MW, 10MW, 20MW]"

    def __len__(self) -> int:
        return len(self.list_scenarios())

    def __iter__(self) -> Generator[ReserveScenario, None, None]:  # type: ignore
        yield from self.list_scenarios()

    def list_scenarios(self) -> List[ReserveScenario]:
        return [
            ReserveScenario(
                volume=volume,
                auction=self.auction,
                product=self.product,
                block=self.block,
                reserve_group=self.reserve_group,
                mip_plant_time_series=self.mip_plant_time_series,
                obligation_external_id=self.obligation_external_id,
            )
            for volume in self.volumes
        ]


class RKOMBidProcessConfig(Configuration):
    watercourse: str = Field(alias="bid_watercourse")

    price_scenarios: List[PriceScenarioID] = Field(alias="bid_price_scenarios")
    reserve_scenarios: ReserveScenarios = Field(alias="bid_reserve_scenarios")

    shop_start: RelativeTime = Field(alias="shop_starttime")
    shop_end: RelativeTime = Field(alias="shop_endtime")

    timezone: str = "Europe/Oslo"
    method: str = "simple"

    minimum_price: int = 0  # TODO: need to specify currency
    price_premium: int = 0  # TODO: need to specify currency

    parent_external_id: typing.ClassVar[str] = "rkom_bid_process_configurations"
    mapping_type: ClassVar[str] = "rkom_incremental_mapping"

    @root_validator(pre=True)
    def create_reserve_scenarios(cls, values):
        if not isinstance(volumes := values.get("bid_reserve_scenarios"), str):
            return values
        volumes = [int(volume.removesuffix("MW")) for volume in volumes[1:-1].split(",")]

        values["bid_reserve_scenarios"] = dict(
            volumes=volumes,
            auction=values["bid_auction"],
            product=values["bid_product"],
            block=values["bid_block"],
            reserve_group=values["labels"][0]["externalId"],
            mip_plant_time_series=[],
        )
        return values

    @validator("shop_start", "shop_end", pre=True)
    def json_loads(cls, value):
        return {"operations": json.loads(value)} if isinstance(value, str) else value

    @validator("price_scenarios", pre=True)
    def literal_eval(cls, value):
        return [{"id": id_} for id_ in ast.literal_eval(value)] if isinstance(value, str) else value

    @property
    def sorted_volumes(self) -> List[int]:
        return sorted(self.reserve_scenarios.volumes)

    @property
    def name(self) -> str:
        return (
            f"{self.watercourse}_"
            f"{self.reserve_scenarios.auction.value}_"
            f"{self.reserve_scenarios.product}_"
            f"{self.reserve_scenarios.block}_"
            f"{len(self.price_scenarios)}-prices_"
            f"{self.sorted_volumes[1]}MW-{self.sorted_volumes[-1]}MW"
        )

    @property
    def external_id(self) -> str:
        return f"POWEROPS_{self.name}"

    @property
    def bid_date(self) -> RelativeTime:
        if self.reserve_scenarios.auction == "week":
            return RelativeTime(relative_time_string="monday")
        else:
            return RelativeTime(relative_time_string="saturday")

    @property
    def rkom_plants(self) -> List[str]:
        return [plant for plant, _ in self.reserve_scenarios.mip_plant_time_series]

    def to_metadata(self, rkom_price_scenarios_by_id: dict[str, PriceScenario], rkom_market_name: str) -> dict:
        price_scenarios = map_price_scenarios_by_name(
            self.price_scenarios, rkom_price_scenarios_by_id, rkom_market_name
        )

        return {
            "bid:watercourse": self.watercourse,
            "bid:auction": self.reserve_scenarios.auction.value,
            "bid:block": self.reserve_scenarios.block,
            "bid:product": self.reserve_scenarios.product,
            "bid:method": self.method,
            "bid:date": str(self.bid_date),
            "bid:price_scenarios": str(list(price_scenarios)),
            "bid:reserve_scenarios": str(self.reserve_scenarios),
            "bid:minimum_price": str(self.minimum_price),
            "bid:price_premium": str(self.price_premium),
            "rkom:plants": str(sorted(self.rkom_plants)),
            "shop:starttime": str(self.shop_start),
            "shop:endtime": str(self.shop_end),
            "timezone": self.timezone,
        }

    def to_cdf_asset(self, rkom_price_scenarios_by_id: dict[str, PriceScenario], rkom_market_name: str) -> Asset:
        return Asset(
            external_id=self.external_id,
            name=self.name,
            metadata=self.to_metadata(rkom_price_scenarios_by_id, rkom_market_name),
            description=f"RKOM bid generation config for {self.watercourse}",
            parent_external_id=self.parent_external_id,
            labels=[Label(AssetLabels.RKOM_BID_CONFIGURATION)],
        )

    def to_bootstrap_resources(self, price_scenarios_by_id, market_name) -> BootstrapResourceCollection:
        bootstrap_resources = BootstrapResourceCollection()
        asset = self.to_cdf_asset(price_scenarios_by_id, market_name)
        bootstrap_resources.add(asset)

        price_scenarios = map_price_scenarios_by_name(self.price_scenarios, price_scenarios_by_id, market_name)

        # Create incremental mapping for each combination of price scenario and reserve_scenario
        for price_scenario_name, price_scenario in price_scenarios.items():
            price_mapping = price_scenario.to_time_series_mapping()
            for reserve_scenario in self.reserve_scenarios:
                reserve_mapping = reserve_scenario.to_time_series_mapping()

                bootstrap_resources += write_mapping_to_sequence(
                    watercourse=self.watercourse,
                    mapping=price_mapping + reserve_mapping,
                    mapping_type=self.mapping_type,  # type: ignore
                    reserve_volume=reserve_scenario.volume,
                    price_scenario_name=price_scenario_name,
                    config_name=self.name,
                )

        for sequence in bootstrap_resources.sequences.values():
            relationship = simple_relationship(
                source=asset,
                target=sequence,
                label_external_id=RelationshipLabels.INCREMENTAL_MAPPING_SEQUENCE,
            )
            bootstrap_resources.add(relationship)

        return bootstrap_resources


ExternalId = str


class PlantTimeSeriesMapping(BaseModel):
    plant_name: str
    water_value: Optional[ExternalId]
    inlet_reservoir_level: Optional[ExternalId]
    outlet_reservoir_level: Optional[ExternalId]
    p_min: Optional[ExternalId]
    p_max: Optional[ExternalId]
    feeding_fee: Optional[ExternalId]
    head_direct: Optional[ExternalId]


class GeneratorTimeSeriesMapping(BaseModel):
    generator_name: str
    start_stop_cost: Optional[ExternalId]


class BootstrapConfig(BaseModel):
    constants: CommonConstants
    cdf: CDFConfig
    benchmarks: list[BenchmarkingConfig]
    price_scenario_by_id: dict[str, PriceScenario]
    bidprocess: list[BidProcessConfig]
    bidmatrix_generators: list[BidMatrixGeneratorConfig]
    dayahead_price_timeseries: Dict[str, str]
    market: MarketConfig
    watercourses: list[WatercourseConfig]
    time_series_mappings: list[TimeSeriesMapping]
    rkom_bid_process: list[RKOMBidProcessConfig]
    rkom_bid_combination: Optional[list[RKOMBidCombinationConfig]] = None
    rkom_market: Optional[RkomMarketConfig] = None
    plant_time_series_mappings: list[PlantTimeSeriesMapping] = None
    generator_time_series_mappings: list[GeneratorTimeSeriesMapping] = None

    @validator("price_scenario_by_id")
    def no_duplicated_scenarios(cls, value: dict[str, PriceScenario]):
        scenarios_by_hash = defaultdict(list)
        for id_, s in value.items():
            scenarios_by_hash[hash(s.json(exclude={"name"}))].append((id_, s))
        if duplicated := [s for s in scenarios_by_hash.values() if len(s) > 1]:
            sep = ") |\n\t\t\t ("
            raise ValueError(
                f"Scenarios  "
                f"({sep.join([', '.join([id_ for id_, _ in duplicate_set]) for duplicate_set in duplicated])}) "
                f"\nare duplicated."
            )
        return value

    @validator("rkom_bid_combination", each_item=True)
    def valid_process_external_id(cls, value, values: dict):
        valid_ids = {process_config.external_id for process_config in values["rkom_bid_process"]}
        for external_id_to_validate in value.rkom_bid_config_external_ids:
            if external_id_to_validate not in valid_ids:
                raise ValueError(
                    f"Reference to rkom bid process config in rkom_bid_combination yaml is wrong for "
                    f"{external_id_to_validate}. "
                    f"Possible references are: {[config.external_id for config in values['rkom_bid_process']]}"
                )
        return value

    @classmethod
    def from_yamls(cls, config_dir_path: Path) -> "BootstrapConfig":
        configs = {}
        for field_name in cls.__fields__:
            if (config_file_path := config_dir_path / f"{field_name}.yaml").exists():
                configs[field_name] = load_yaml(config_file_path, encoding="utf-8")
        return cls(**configs)

    def validate_bid_configs(self):
        """Validate the bid configs in the bootstrap config. Per now only ensure there is at most one default config
        per price area"""

        default_configs = defaultdict(int)
        for bid_config in self.bidprocess:
            if bid_config.is_default_config_for_price_area:
                default_configs[bid_config.price_area_name] += 1

        if price_areas_with_multiple_default_configs := [
            price_area for price_area, count in default_configs.items() if count > 1
        ]:
            raise ValueError(
                f"Multiple default bid configs for price areas: {price_areas_with_multiple_default_configs}"
            )



"""
This file is auto-generated using `dm topython` CLI tool.
"""

from __future__ import annotations

import logging
import sys
from typing import List, Optional

from cognite.dm_clients.domain_modeling import DomainModel, Schema

logger = logging.getLogger(__name__)

power_ops_dm_schema: Schema[DomainModel] = Schema()


@power_ops_dm_schema.register_type
class ProcessingLog(DomainModel):
    state: Optional[str] = None
    timestamp: Optional[str] = None
    error_msg: Optional[str] = None


@power_ops_dm_schema.register_type(lowercase_type_name="commands")
class CommandsConfig(DomainModel):
    commands: List[str]


@power_ops_dm_schema.register_type
class FileRef(DomainModel):
    type: str
    file_external_id: str


@power_ops_dm_schema.register_type
class Transformation(DomainModel):
    method: str
    arguments: str


@power_ops_dm_schema.register_type
class Mapping(DomainModel):
    path: str
    timeseries_external_id: Optional[str] = None
    transformations: Optional[List[Optional[Transformation]]] = []
    retrieve: Optional[str] = None
    aggregation: Optional[str] = None


@power_ops_dm_schema.register_type
class ModelTemplate(DomainModel):
    version: str
    shop_version: str
    watercourse: str
    model: Optional[FileRef] = None
    base_mappings: Optional[List[Optional[Mapping]]] = []


@power_ops_dm_schema.register_type
class Scenario(DomainModel):
    name: str
    model_template: Optional[ModelTemplate] = None
    mappings_override: Optional[List[Optional[Mapping]]] = []
    commands: Optional[CommandsConfig] = None
    extra_files: Optional[List[Optional[FileRef]]] = []


@power_ops_dm_schema.register_type(root_type=True)
class Case(DomainModel):
    scenario: Optional[Scenario] = None
    start_time: str
    end_time: str
    processing_log: Optional[List[Optional[ProcessingLog]]] = []


# Keep at the end of file:
power_ops_dm_schema.close()


# Render the schema to stdout when executed directly:
if __name__ == "__main__":
    sys.stdout.write(power_ops_dm_schema.as_str())



import traceback
from enum import Enum, auto
from functools import wraps
from typing import Callable, Optional
from uuid import uuid4

from cognite.client import CogniteClient
from cognite.client.data_classes import Event
from cognite.client.exceptions import CogniteAPIError, CogniteMultiException
from retry import retry

from cognite.powerops.preprocessor import knockoff_logging as logging
from cognite.powerops.preprocessor.config import RelationshipsConfig, StatusEventConfig
from cognite.powerops.preprocessor.utils import create_relationship, now

logger = logging.getLogger(__name__)


class EventStatus(Enum):
    STARTED = auto()
    FINISHED = auto()
    FAILED = auto()


def _create_status_event(
    client: CogniteClient,
    status: EventStatus,
    parent_event: Event,
    extra_metadata: Optional[dict] = None,
) -> Event:
    suffix = str(uuid4())
    current_time = now()
    external_id = f"{StatusEventConfig.EVENT_EXTERNAL_ID_PREFIX}_{status.name}_{suffix}"
    event_type = f"{StatusEventConfig.EVENT_TYPE_PREFIX}_{status.name}"
    metadata = {"event_external_id": parent_event.external_id}
    if extra_metadata:
        metadata = {**extra_metadata, **metadata}
    event = Event(
        start_time=current_time,
        end_time=current_time,
        external_id=external_id,
        type=event_type,
        subtype=None,
        source=StatusEventConfig.EVENT_SOURCE,
        data_set_id=parent_event.data_set_id,
        metadata=metadata,
    )
    client.events.create(event=event)
    logger.info(f"Created status event '{event.external_id}'")
    return event


@retry(
    (CogniteAPIError, CogniteMultiException),
    tries=6,
    delay=0.5,
    backoff=1.2,  # increasing delays from 500ms to about 1s (500ms * 1.2^(6-1))
)
def _post_status_event(
    client: CogniteClient,
    status: EventStatus,
    parent_event_external_id: str,
    extra_metadata: Optional[dict] = None,
) -> None:
    """Posts a status event to CDF that is connected to the parent event through a relationship."""
    parent_event = client.events.retrieve(external_id=parent_event_external_id)
    event = _create_status_event(client=client, status=status, parent_event=parent_event, extra_metadata=extra_metadata)
    label_external_id = f"{RelationshipsConfig.LABEL_PREFIX}status_event_{status.name.lower()}"
    relationship = create_relationship(
        client=client,
        source=parent_event,
        target=event,
        label_external_id=label_external_id,
        data_set_id=parent_event.data_set_id,
    )
    logger.info(
        f"Status event '{event.external_id}' connected to '{parent_event.external_id}' by relationship"
        f" '{relationship.external_id}'"
    )


def wrap_status_events(func: Callable) -> Callable:
    """Wrapper that handles writing status events to CDF"""

    @wraps(func)
    def wrapper(client: CogniteClient, event_external_id: str):
        try:
            _post_status_event(
                client=client,
                status=EventStatus.STARTED,
                parent_event_external_id=event_external_id,
            )
            res = func(client, event_external_id)
            _post_status_event(
                client=client,
                status=EventStatus.FINISHED,
                parent_event_external_id=event_external_id,
            )
            return res
        except Exception:
            logger.error("Something went wrong:")
            stack_trace = traceback.format_exc()
            logger.error(stack_trace)
            extra_metadata = {
                "errorStackTrace": stack_trace,
            }  # "errorStackTrace" used by front end
            _post_status_event(
                client=client,
                status=EventStatus.FAILED,
                parent_event_external_id=event_external_id,
                extra_metadata=extra_metadata,
            )
            raise

    return wrapper



from typing import Callable

from cognite.client.data_classes import LabelDefinition


# All Labels used by PowerOps
class AssetLabels:
    BENCHMARKING_CONFIGURATION = "benchmarking_configuration"
    BID_PROCESS_CONFIGURATION = "bid_process_configuration"
    BID_MATRIX_GENERATOR_CONFIG_SEQUENCE = "bid_matrix_generator_config_sequence"
    INCREMENTAL_MAPPING_SEQUENCE = "incremental_mapping_sequence"
    CONNECTION_BYPASS = "connection_bypass"
    CONNECTION_SPILL = "connection_spill"
    CONNECTION_STANDARD = "connection_standard"
    DAYAHEAD_BIDDING_BENCHMARKING_CONFIG = "dayahead_bidding_benchmarking_config"
    GATE = "gate"
    GATE_RESERVOIR = "gate_reservoir"
    GENERATOR = "generator"
    GENERATOR_PLANT = "generator_plant"
    MARKET = "market"
    PLANT = "plant"
    PLANT_RESERVOIR = "plant_reservoir"
    PRICE_AREA = "price_area"
    RESERVE_GROUP = "reserve_group"
    RESERVE_GROUP_GENERATOR = "reserve_group_generator"
    RESERVOIR = "reservoir"
    RESERVOIR_GATE = "reservoir_gate"
    RESERVOIR_PLANT = "reservoir_plant"
    RKOM_BID_CONFIGURATION = "rkom_bid_configuration"
    WATERCOURSE = "watercourse"


class RelationshipLabels:
    BID_MATRIX_GENERATOR_CONFIG_SEQUENCE = "relationship_to.bid_matrix_generator_config_sequence"
    BID_MATRIX_SEQUENCE = "relationship_to.bid_matrix_sequence"
    RKOM_BID_FORM_SEQUENCE = "relationship_to.rkom_bid_form_sequence"
    BID_PROCESS_CONFIGURATION_ASSET = "relationship_to.bid_process_configuration_asset"
    BID_PROCESS_EVENT = "relationship_to.bid_process_event"
    CALCULATE_TOTAL_BID_MATRIX_EVENT = "relationship_to.calculate_total_bid_matrix_event"
    DAYAHEAD_PRICE_TIME_SERIES = "relationship_to.dayahead_price_time_series"
    GATE_DISCHARGE_TIME_SERIES = "relationship_to.gate_discharge_time_series"
    GENERATOR_DISCHARGE_TIME_SERIES = "relationship_to.generator_discharge_time_series"
    GENERATOR_PRODUCTION_TIME_SERIES = "relationship_to.generator_production_time_series"
    INCREMENTAL_MAPPING_SEQUENCE = "relationship_to.incremental_mapping_sequence"
    LOG_FILE = "relationship_to.log_file"
    MARKET_ASSET = "relationship_to.market_asset"
    MARKET_PRICE_TIME_SERIES = "relationship_to.market_price_time_series"
    MARKET_SALES_TIME_SERIES = "relationship_to.market_sales_time_series"
    OBJECTIVE_SEQUENCE = "relationship_to.objective_sequence"
    PARENT_COLLECTION = "relationship_to.parent_collection"
    PARENT_PROCESS = "relationship_to.parent_process"
    PLANT = "relationship_to.plant"
    PLANT_DISCHARGE_TIME_SERIES = "relationship_to.plant_discharge_time_series"
    PLANT_PRODUCTION_TIME_SERIES = "relationship_to.plant_production_time_series"
    PLANT_CONSUMPTION_TIME_SERIES = "relationship_to.plant_consumption_time_series"
    RESERVOIR_WATER_VALUE_TIME_SERIES = "relationship_to.reservoir_water_value_time_series"
    RESERVOIR_ENERGY_CONVERSION_FACTOR_SERIES = "relationship_to.reservoir_energy_conversion_factor_time_series"
    PRODUCTION_OBLIGATION_TIME_SERIES = "relationship_to.production_obligation_time_series"
    PRODUCTION_SCHEDULE = "relationship_to.production_schedule"
    CONSUMPTION_SCHEDULE = "relationship_to.consumption_schedule"
    RESERVOIR_LEVEL_TIME_SERIES = "relationship_to.reservoir_level_time_series"
    RESERVOIR_VOLUME_TIME_SERIES = "relationship_to.reservoir_volume_time_series"
    RKOM_COLLECTION_CONFIG = "relationship_to.rkom_collection_config"
    SHOP_RUN_EVENT = "relationship_to.shop_run_event"
    STATUS_EVENT_STARTED = "relationship_to.status_event_started"
    STATUS_EVENT_FINISHED = "relationship_to.status_event_finished"
    STATUS_EVENT_FAILED = "relationship_to.status_event_failed"
    TRIGGER_EVENT = "relationship_to.trigger_event"
    WATERCOURSE = "relationship_to.watercourse"
    INLET_RESERVOIR = "relationship_to.inlet_reservoir"
    GENERATOR = "relationship_to.generator"
    GENERATOR_EFFICIENCY_CURVE = "relationship_to.generator_efficiency_curve"
    TURBINE_EFFICIENCY_CURVE = "relationship_to.turbine_efficiency_curve"
    P_MIN_TIME_SERIES = "relationship_to.p_min_time_series"
    P_MAX_TIME_SERIES = "relationship_to.p_max_time_series"
    WATER_VALUE_TIME_SERIES = "relationship_to.water_value_time_series"
    FEEDING_FEE_TIME_SERIES = "relationship_to.feeding_fee_time_series"
    INLET_LEVEL_TIME_SERIES = "relationship_to.inlet_level_time_series"
    OUTLET_LEVEL_TIME_SERIES = "relationship_to.outlet_level_time_series"
    HEAD_DIRECT_TIME_SERIES = "relationship_to.head_direct_time_series"
    START_STOP_COST_TIME_SERIES = "relationship_to.start_stop_cost_time_series"
    CASE_FILE = "relationship_to.case_file"
    CUT_FILE = "relationship_to.cut_file"
    MAPPING_FILE = "relationship_to.mapping_file"
    EXTRA_FILE = "relationship_to.extra_file"


def label_external_ids(labels_class: Callable) -> list:
    return [v for k, v in labels_class.__dict__.items() if not k.startswith("__")]


def create_labels() -> list[LabelDefinition]:
    external_ids = label_external_ids(AssetLabels) + label_external_ids(RelationshipLabels)
    return [
        LabelDefinition(
            external_id=label_external_id,
            name=label_external_id,
        )
        for label_external_id in external_ids
    ]



from __future__ import annotations

import logging
import time
from enum import Enum
from typing import TYPE_CHECKING

from cognite.powerops.client.data_classes.shop_results import ShopRunResult
from cognite.powerops.client.data_classes.shop_run_event import ShopRunEvent

if TYPE_CHECKING:
    from cognite.powerops import PowerOpsClient

logger = logging.getLogger(__name__)


class ShopRun:
    class Status(Enum):
        IN_PROGRESS = "IN_PROGRESS"
        SUCCEEDED = "SUCCEEDED"
        FAILED = "FAILED"

    def __init__(
        self,
        po_client: PowerOpsClient,
        *,
        shop_run_event: ShopRunEvent,
    ) -> None:
        self._po_client = po_client
        self.shop_run_event = shop_run_event

    @property
    def in_progress(self) -> bool:
        return self.status == ShopRun.Status.IN_PROGRESS

    @property
    def succeeded(self) -> bool:
        return self.status == ShopRun.Status.SUCCEEDED

    @property
    def failed(self) -> bool:
        return self.status == ShopRun.Status.FAILED

    @property
    def status(self) -> ShopRun.Status:
        return self._po_client.shop.runs.retrieve_status(self.shop_run_event.external_id)

    def wait_until_complete(self) -> None:
        while self.in_progress:
            logger.debug(f"{self.shop_run_event.external_id} is still running...")
            time.sleep(3)
        logger.debug(f"{self.shop_run_event.external_id} finished.")

    def get_results(self, wait: bool = True) -> ShopRunResult:
        if wait:
            if self.in_progress:
                logger.warning(f"{self.shop_run_event.external_id} is still running, waiting for results...")
            self.wait_until_complete()
        return self._po_client.shop.results.retrieve(self)

    def __repr__(self) -> str:
        return f'<ShopRun status="{self.status}" event_external_id="{self.shop_run_event.external_id}">'

    def __str__(self) -> str:
        return self.__repr__()



from __future__ import annotations

from cognite.client.data_classes import Asset, Label, Relationship, Sequence

from cognite.powerops.utils.labels import RelationshipLabels


def basic_relationship(
    source_external_id: str,
    source_type: str,
    target_external_id: str,
    target_type: str,
    label: str,
) -> Relationship:
    return Relationship(
        external_id=f"{source_external_id}.{target_external_id}",
        source_external_id=source_external_id,
        source_type=source_type,
        target_external_id=target_external_id,
        target_type=target_type,
        labels=[Label(external_id=label)],
    )


def asset_to_asset(
    source: Asset | str,
    target: Asset | str,
    label: str,
) -> Relationship:
    return basic_relationship(
        source_type="ASSET",
        target_type="ASSET",
        source_external_id=source if isinstance(source, str) else source.external_id,
        target_external_id=target if isinstance(target, str) else target.external_id,
        label=label,
    )


def asset_to_time_series(
    source_external_id: str,
    target_external_id: str,
    label: str,
) -> Relationship:
    return basic_relationship(
        source_type="ASSET",
        target_type="TIMESERIES",
        source_external_id=source_external_id,
        target_external_id=target_external_id,
        label=label,
    )


def _asset_to_time_series(
    source_external_id: str,
    target_external_id: str,
    label: str,
) -> Relationship:
    return asset_to_time_series(
        source_external_id=source_external_id,
        target_external_id=target_external_id,
        label=label,
    )


def _asset_to_sequence(
    source_external_id: str,
    target_external_id: str,
    label: str,
) -> Relationship:
    return Relationship(
        external_id=f"{source_external_id}.{target_external_id}",
        source_external_id=source_external_id,
        source_type="ASSET",
        target_external_id=target_external_id,
        target_type="SEQUENCE",
        labels=[Label(external_id=label)],
    )


def price_area_to_plant(price_area: Asset | str, plant: Asset | str) -> Relationship:
    return asset_to_asset(
        source=price_area,
        target=plant,
        label=RelationshipLabels.PLANT,
    )


def watercourse_to_plant(watercourse: Asset | str, plant: Asset | str) -> Relationship:
    return asset_to_asset(
        source=watercourse,
        target=plant,
        label=RelationshipLabels.PLANT,
    )


def price_area_to_watercourse(price_area: Asset | str, watercourse: Asset | str) -> Relationship:
    return asset_to_asset(
        source=price_area,
        target=watercourse,
        label=RelationshipLabels.WATERCOURSE,
    )


def plant_to_inlet_reservoir(plant: Asset | str, reservoir: Asset | str) -> Relationship:
    return asset_to_asset(
        source=plant,
        target=reservoir,
        label=RelationshipLabels.INLET_RESERVOIR,
    )


def plant_to_generator(plant: Asset | str, generator: Asset | str) -> Relationship:
    return asset_to_asset(
        source=plant,
        target=generator,
        label=RelationshipLabels.GENERATOR,
    )


def watercourse_to_production_obligation(watercourse: Asset, production_obligation_ts_ext_id: str) -> Relationship:
    return _asset_to_time_series(
        source_external_id=watercourse.external_id,
        target_external_id=production_obligation_ts_ext_id,
        label=RelationshipLabels.PRODUCTION_OBLIGATION_TIME_SERIES,
    )


def generator_to_generator_efficiency_curve(generator: Asset, generator_efficiency_curve: Sequence) -> Relationship:
    return _asset_to_sequence(
        source_external_id=generator.external_id,
        target_external_id=generator_efficiency_curve.external_id,
        label=RelationshipLabels.GENERATOR_EFFICIENCY_CURVE,
    )


def generator_to_turbine_efficiency_curve(generator: Asset, turbine_efficiency_curve: Sequence) -> Relationship:
    return _asset_to_sequence(
        source_external_id=generator.external_id,
        target_external_id=turbine_efficiency_curve.external_id,
        label=RelationshipLabels.TURBINE_EFFICIENCY_CURVE,
    )


def price_area_to_dayahead_price(price_area_asset: Asset, dayahead_price_time_series_external_id: str) -> Relationship:
    return _asset_to_time_series(
        source_external_id=price_area_asset.external_id,
        target_external_id=dayahead_price_time_series_external_id,
        label=RelationshipLabels.DAYAHEAD_PRICE_TIME_SERIES,
    )



from cognite.powerops.config import (
    BenchmarkingConfig,
    BidProcessConfig,
    MarketConfig,
    RKOMBidCombinationConfig,
    RKOMBidProcessConfig,
    Watercourse,
)
from cognite.powerops.data_classes.core import AssetResourceList
from cognite.powerops.data_classes.generator import Generator
from cognite.powerops.data_classes.plant import Plant
from cognite.powerops.data_classes.price_area import PriceArea
from cognite.powerops.data_classes.reservoirs import Reservoir


class WatercourseList(AssetResourceList):
    _RESOURCE = Watercourse


class PlantList(AssetResourceList):
    _RESOURCE = Plant


class ReservoirList(AssetResourceList):
    _RESOURCE = Reservoir


class GeneratorList(AssetResourceList):
    _RESOURCE = Generator


class PriceAreaList(AssetResourceList):
    _RESOURCE = PriceArea


class BidProcessList(AssetResourceList):
    _RESOURCE = BidProcessConfig


class RKOMBidList(AssetResourceList):
    _RESOURCE = RKOMBidProcessConfig


class BenchmarkingList(AssetResourceList):
    _RESOURCE = BenchmarkingConfig


class MarketConfigsList(AssetResourceList):
    _RESOURCE = MarketConfig


class RKOMBidCombinationList(AssetResourceList):
    _RESOURCE = RKOMBidCombinationConfig



import re
from typing import Optional

from cognite.client.data_classes import Asset, Label

from cognite.powerops.utils.labels import AssetLabels


def price_area_asset(name: str) -> Asset:
    return Asset(
        external_id=f"price_area_{name}",
        name=name,
        parent_external_id="price_areas",
        labels=[Label(AssetLabels.PRICE_AREA)],
    )


def watercourse_asset(name: str, shop_penalty_limit: float) -> Asset:
    return Asset(
        external_id=f"watercourse_{name}",
        name=name,
        parent_external_id="watercourses",
        labels=[Label(AssetLabels.WATERCOURSE)],
        metadata={"shop:penalty_limit": str(shop_penalty_limit)},
    )


def plant_asset(name: str, display_name: Optional[str] = None, ordering_key: Optional[float] = None) -> Asset:
    return Asset(
        external_id=f"plant_{name}",
        name=name,
        parent_external_id="plants",
        metadata={
            "display_name": display_name or re.sub(r"\([0-9]+\)", "", name),
            "ordering": str(ordering_key) if ordering_key else "999",
        },
        labels=[Label(AssetLabels.PLANT)],
    )


def reservoir_asset(
    name: str,
    display_name: Optional[str] = None,
    ordering_key: Optional[float] = None,
) -> Asset:
    return Asset(
        external_id=f"reservoir_{name}",
        name=name,
        parent_external_id="reservoirs",
        metadata={
            "display_name": display_name or re.sub(r"\([0-9]+\)", "", name),
            "ordering": str(ordering_key) if ordering_key else "999",
        },
        labels=[Label(AssetLabels.RESERVOIR)],
    )


def generator_asset(name: str, penstock: str = "1", startcost: float = 0) -> Asset:
    return Asset(
        external_id=f"generator_{name}",
        name=name,
        parent_external_id="generators",
        labels=[Label(AssetLabels.GENERATOR)],
        metadata={"penstock": penstock, "startcost": startcost},
    )



import os

from cognite.client import CogniteClient, config
from cognite.client.credentials import OAuthClientCredentials, OAuthDeviceCode


def get_client(parameters: dict) -> CogniteClient:
    client_name = "power-ops-bootstrap"

    client_id = parameters.get("CLIENT_ID")
    cluster = parameters.get("CDF_CLUSTER")
    project = parameters.get("COGNITE_PROJECT")
    tenant_id = parameters.get("TENANT_ID")

    base_url = f"https://{cluster}.cognitedata.com"
    scopes = [f"https://{cluster}.cognitedata.com/.default"]

    if client_secret := os.getenv("POWER_OPS_CLIENT_SECRET"):
        token_url = f"https://login.microsoftonline.com/{tenant_id}/oauth2/v2.0/token"
        creds = OAuthClientCredentials(
            token_url=token_url,
            client_id=client_id,
            client_secret=client_secret,
            scopes=scopes,
        )

    else:
        authority_host_uri = "https://login.microsoftonline.com"
        authority_uri = f"{authority_host_uri}/{tenant_id}"

        creds = OAuthDeviceCode(
            authority_url=authority_uri,
            client_id=client_id,
            scopes=scopes,
        )

    cnf = config.ClientConfig(
        credentials=creds,
        project=project,
        base_url=base_url,
        client_name=client_name,
    )

    return CogniteClient(cnf)



from __future__ import annotations

import logging
from datetime import datetime
from typing import Optional, Sequence

import matplotlib.pyplot as plt

from cognite.powerops.client.data_classes.shop_result_files import ShopYamlFile
from cognite.powerops.utils.plotting import ax_plot_time_series, create_time_series_plot

logger = logging.getLogger(__name__)


class ShopResultsCompare:
    def plot_time_series(
        self,
        post_run_list: Sequence[ShopYamlFile],
        comparison_key: str,
        labels: Optional[Sequence[str]] = None,
    ):
        """Stacked line charts of the given post runs at the same key.
        Labels must be in the same order as the post runs."""
        if labels and (len(labels) != len(set(labels)) or len(labels) != len(post_run_list)):
            logger.error("Titles must be unique and match the number of provided post runs")
            return

        plots: dict[str, dict[datetime, float]] = {}

        for i, shop_yaml in enumerate(post_run_list):
            if time_series := shop_yaml._retrieve_time_series_dict(comparison_key):
                plot_label = labels[i] if labels else shop_yaml.name
                plots[plot_label] = time_series

        ax = create_time_series_plot()
        for label, time_series in plots.items():
            ax_plot_time_series(ax, time_series, label)
        ax.legend()
        plt.show()

    def yaml_difference(
        self,
        yaml_1: ShopYamlFile,
        yaml_2: ShopYamlFile,
    ):
        raise NotImplementedError()
